/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: atreides.muaddib@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/

#include "nSX2_type.h"

#ifndef _DMAC_H
#define _DMAC_H


/* Define or Undefine if you want DMAC LOG */
#define	DMAC_LOG	DMAC_Log
//#undef	DMAC_LOG



/********************************************************************************/
/*						DMA CONTROLLER DEFINES									*/
/********************************************************************************/



/* DMA CHANNEL DEFINES */
#define DMA_CH0 	0		//Linked to VIF0
#define DMA_CH1 	1		//Linked to VIF1
#define DMA_CH2 	2		//Linked to GIF
#define DMA_CH3 	3		//Linked to IPU (FROM???)
#define DMA_CH4 	4		//Linked to IPU (TO???)
#define DMA_CH5 	5		//Linked to SIF0
#define DMA_CH6 	6		//Linked to SIF1
#define DMA_CH7 	7		//Linked to SIF2
#define DMA_CH8 	8		//Linked to SPR  (from SCRATCH PAD to RAM???)
#define DMA_CH9 	9		//Linked to SPR  (from RAM to SCRATCH PAD???)

/* NUMBER MAX OF DMA CHANNEL */
#define	DMA_CHANNEL_NUMBER	10

/* DMAC REGISTERS */
#define DMAC_CTRL    0x1000E000          //DMAC Control
#define DMAC_STAT    0x1000E010          //DMAC Status
#define DMAC_PCR     0x1000E020          //DMAC Priority control
#define DMAC_SQWC    0x1000E030          //DMAC skip quad word
#define DMAC_RBSR    0x1000E040          //DMAC ring buffer size
#define DMAC_RBOR    0x1000E050          //DMAC Ring buffer offset
#define DMAC_STADR   0x1000E060          //DMA stall address


/********************************************************************************/
/*						DMA CHANNELS DEFINES									*/
/********************************************************************************/



/* DMA CH0 REGISTERS (Linked to VIF0) */
#define D0_CHCR 	0x10008000		//Ch0 Channel Control
#define	D0_MADR  	0x10008010		//Ch0 Memory Address
#define D0_QWC 		0x10008020 		//Ch0 Quad word count
#define D0_TADG		0x10008030		//Ch0 Tag address
#define D0_ASR0		0x10008040		//CH0 Address stack 0
#define D0_ASR1		0x10008050		//Ch0 Address stack 1
#define D0_SADR		0x10008080		//Ch0 

/* DMA CH1 REGISTERS (Linked to VIF1) */
#define D1_CHCR 	0x10009000		//Ch1 Channel Control
#define D1_MADR 	0x10009010		//Ch1 Memory Address
#define D1_QWC		0x10009020		//Ch1 Quad word count
#define D1_TADR		0x10009030		//Ch1 Tag address
#define D1_ASR0		0x10009040		//CH1 Address stack 0
#define D1_ASR1		0x10009050		//CH1 Address stack 1
#define D1_SADR		0x10009080		//CH1 

/* DMA CH2 REGISTERS (Linked to GIF) */
#define D2_CHCR 	0x1000A000		//Ch2 Channel Control
#define D2_MADR 	0x1000A010		//Ch2 Memory Address
#define D2_QWC		0x1000A020		//Ch2 Quad word count
#define D2_TADR		0x1000A030		//Ch2 Tag address
#define D2_ASR0		0x1000A040		//CH2 Address stack 0
#define D2_ASR1		0x1000A050		//CH2 Address stack 1
#define D2_SADR		0x1000A080		//CH2 SPR ( meet in Starsim )

/* DMA CH3 REGISTERS (Linked to IPU (FROM IPU)) */
#define D3_CHCR 	0x1000B000		//Ch3 Channel Control
#define D3_MADR 	0x1000B010		//Ch3 Memory Address
#define D3_QWC		0x1000B020		//Ch3 Quad word count
#define D3_TADR 	0x1000B030		//Ch3 
#define D3_ASR0 	0x1000B040		//Ch3 
#define D3_ASR1		0x1000B050		//Ch3 
#define D3_SADR		0x1000B080		//Ch3 

/* DMA CH4 REGISTERS (Linked to IPU (TO IPU)) */
#define D4_CHCR 	0x1000B400		//Ch4 Channel Control
#define D4_MADR 	0x1000B410		//Ch4 Memory Address
#define D4_QWC		0x1000B420		//Ch4 Quad word count
#define D4_TADR 	0x1000B430		//Ch4 Tag address
#define D4_ASR0 	0x1000B440		//Ch4 
#define D4_ASR1		0x1000B450		//Ch4 
#define D4_SADR 	0x1000B480		//Ch4 


/* DMA CH5 REGISTERS (Linked to SIF0) */
#define D5_CHCR 	0x1000C000		//Ch5 Channel Control
#define D5_MADR 	0x1000C010		//Ch5 Memory Address
#define D5_QWC		0x1000C020		//Ch5 Quad word count


/* DMA CH6 REGISTERS (Linked to SIF1) */
#define D6_CHCR 	0x1000C400		//Ch6 Channel Control
#define D6_MADR 	0x1000C410		//Ch6 Memory Address
#define D6_QWC		0x1000C420		//Ch6 Quad word count
#define D6_TADR 	0x1000C430		//Ch6 Tag address


/* DMA CH7 REGISTERS (Linked to SIF2) */
#define D7_CHCR 	0x1000C800		//Ch7 Channel Control
#define D7_MADR 	0x1000C810		//Ch7 Memory Address
#define D7_QWC		0x1000C820		//Ch7 Quad word count


/* DMA CH8 REGISTERS (Linked to SPR  (from SCRATCH PAD) */
#define D8_CHCR 	0x1000D000		//Ch8 Channel Control
#define D8_MADR 	0x1000D010		//Ch8 Memory Address
#define D8_QWC		0x1000D020		//Ch8 Quad word count
#define D8_TADR 	0x1000D030		//Ch8 
#define D8_ASR0 	0x1000D040		//Ch8 
#define D8_ASR1		0x1000D050		//Ch8 
#define D8_SADR		0x1000D080		//Ch8 SPR address


/* DMA CH9 REGISTERS (Linked to SPR  (To SCRATCH PAD) */
#define D9_CHCR 	0x1000D400		//Ch9 Channel Control
#define D9_MADR 	0x1000D410		//Ch9 Memory Address
#define D9_QWC		0x1000D420		//Ch9 quad word count
#define D9_TADR 	0x1000D430		//Ch9 Tag address
#define D9_ASR0		0x1000D440		//Ch9 quad word count
#define D9_ASR1 	0x1000D450		//Ch9 Tag address
#define D9_SADR		0x1000D480		//Ch9 SPR address


/********************************************************************************/
/*						CHAIN SOURCE MODE DEFINES								*/
/********************************************************************************/


/* TAG SOURCE CHAIN MODE : FLAG */
#define	DMA_SRC_CHAIN_MODE_REFE		0x0
#define	DMA_SRC_CHAIN_MODE_CONT		0x1
#define	DMA_SRC_CHAIN_MODE_NEXT		0x2
#define	DMA_SRC_CHAIN_MODE_REF		0x3
#define	DMA_SRC_CHAIN_MODE_REFS		0x4
#define	DMA_SRC_CHAIN_MODE_CALL		0x5
#define	DMA_SRC_CHAIN_MODE_RET		0x6
#define	DMA_SRC_CHAIN_MODE_END		0x7


/* Dma Transfer Size in Byte ( 16 = 128 bits ) */
#define 	DMA_TRS_TAG_SIZE    		16


/* In Source Chain Mode Enable/Disable the TAG Transfert */
#define		ENABLE_TRANSFERT_TAG	TRUE
#define		DISABLE_TRANSFERT_TAG	FALSE

/* In Source Chain Mode Enable/Disable Interrupt */
#define		ENABLE_TAG_IRQ			TRUE
#define		DISABLE_TAG_IRQ			FALSE


/********************************************************************************/
/*						DMA CHANNEL REGISTERS STRUCTURE							*/
/********************************************************************************/


typedef struct _DmaChannelRegisters {
	_u32	ctrl;			// 0x00  -  CHANNEL CONTROL REGISTER 
	_u32	null0[3];		// 0x04 0x08 0x0C
	_u32	memAdd;			// 0x10  -  MEMORY ADRESS
	_u32	null1[3];		// 0x14 0x18 0x1C
	_u32	size;			// 0x20  -  QUADRA-WORD COUNTER
	_u32	null2[3];		// 0x24 0x28 0x2C
	_u32	tag;			// 0x30  -  TAG ADDRESS
	_u32	null3[3];		// 0x34 0x38 0x3C
	_u32	asr0;			// 0x40  -  ADDRESS STACK 0 
	_u32	null4[3];		// 0x44 0x48 0x4C
	_u32	asr1;			// 0x50  -  ADDRESS STACK 1
	_u32	null5[11];		// 0x54 0x58 0x5C 0x60 0x64 0x68 0x6C 0x70 0x74 0x78 0x7C
	_u32	saddr;			// 0x80  -  SPR ADDRESS
} DmaChannelRegisters;




/********************************************************************************/
/*						DMA CONTROLLER REGISTERS STRUCTURE						*/
/********************************************************************************/


typedef struct _DmaControllerRegisters {

	_u32	ctrl;		// 0x00		DMAC Control
	_u32	null0[3];	// 0x04 0x08 0x0C
	_u32	stat;		// 0x10		DMAC Status
	_u32	null1[3];	// 0x14 0x18 0x1C
	_u32	pcr;		// 0x20 	DMAC Priority control
	_u32	null2[3];	// 0x24 0x28 0x2C
	_u32	sqwc;		// 0x30		DMAC skip quad word
	_u32	null3[3];	// 0x34 0x38 0x3C
	_u32	rbsr;		// 0x40		DMAC ring buffer size
	_u32	null4[3];	// 0x44 0x48 0x4C
	_u32	rbor;		// 0x50		DMAC Ring buffer offset
	_u32	null5[3];	// 0x54 0x58 0x5C
	_u32	stadr;		// 0x60		DMA stall address

} DmaControllerRegisters;



/********************************************************************************/
/*						DMA CONTROLLER REGISTERS STRUCTURE						*/
/********************************************************************************/

typedef struct _DMAController {

	
	DmaControllerRegisters *Reg ;						/* DMAC REGISTERS */
	DmaChannelRegisters *Channel[ DMA_CHANNEL_NUMBER ];	/* DMA CHANNELS*/

	
	_u8		EnableAllDMAs;		/* DmaC enable */

}DMAController;
 

extern		DMAController DmaC;


/********************************************************************************/
/*						DMA CONTROLLER  FUNCTIONS								*/
/********************************************************************************/

/* Main */
void 	DMAC_Log 	( char * ftext, ...   );
void	DMAC_Init	( void );
void	DMAC_Write 	( _u64 fPhysicalAddress, _u128 fVal, 	_u8 fDataType );
void	DMAC_Read 	( _u64 fPhysicalAddress, _u128 * fVal, 	_u8 fDataType );


/* SourceChainMode */
void 	DMAC_Transfer_SourceChain_Mode		( 	_u8 Channel, 
												_u8 TransfertTag, _u8 EnableIRQ,
												void (*fNormalMode)(void) );

/* Ch0 VIF0 */
void 	DMAC_VIF0_Transfer 					( void );
void 	DMAC_VIF0_Transfer_Normal_Mode		( void );

/* Ch1 VIF1 */
void 	DMAC_VIF1_Transfer 					( void );
void 	DMAC_VIF1_Transfer_Normal_Mode		( void );


/* Ch2 GIF */
void 	DMAC_GIF_Transfer 					( void );
void 	DMAC_GIF_Transfer_Normal_Mode		( void );


/* Ch3 FromIPU */
void 	DMAC_FromIPU_Transfer 				( void );


/* Ch4 ToIPU */
void 	DMAC_ToIPU_Transfer 				( void );


/* Ch5 SIF0 */
void 	DMAC_SIF0_Transfer 					( void );


/* Ch6 SIF1 */
void 	DMAC_SIF1_Transfer 					( void );


/* Ch7 SIF2 */
void 	DMAC_SIF2_Transfer 					( void );

/* Ch8 FromSPR */
void 	DMA_FromSPR_Transfer 				( void );
void 	DMAC_FromSPR_Transfer_Normal_Mode	( void );

/* Ch9 ToSPR */
void 	DMAC_ToSPR_Transfer 				( void );
void 	DMAC_ToSPR_Transfer_Normal_Mode		( void );


#endif


