/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: atreides.muaddib@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/

#ifndef _GIF_H
#define _GIF_H



/* Define or Undefine if you want GIF LOG */
/* GIF REGISTERS */
#define	GIF_LOG	GIF_Log
//#undef	GIF_LOG

/* GIF STATUS */
#define	GIF_LOG_ST	GIF_Log
//#undef	GIF_LOG_ST


/********************************************************************************************
*					GIF REGISTERS DEFINES													*		
*********************************************************************************************/

#define	GIF_CTRL	0x10003000        	//Control
#define	GIF_MODE	0x10003010        	//Mode setting
#define	GIF_STAT	0x10003020        	//Status
#define	GIF_TAG0	0x10003040        	//GIFtag (bits 31-0) immediately before
#define	GIF_TAG1	0x10003050        	//GIFtag (bits 63-32) immediately before
#define	GIF_TAG2	0x10003060        	//GIFtag (bits 95-64) immediately before
#define	GIF_TAG3	0x10003070        	//GIFtag (bits 127-96) immediately before
#define	GIF_CNT		0x10003080        	//Transfer status counter
#define	GIF_P3CNT	0x10003090        	//PATH3 transfer status counter
#define	GIF_P3TAG	0x100030a0        	//PATH3 tag value



/********************************************************************************************
*					GIF REGISTERS STRUCTURE													*		
*********************************************************************************************/

typedef struct _GIFRegisters {

	_u32	ctrl;		// 0x00	- Control	
	_u32	null0[3];	// 0x04 0x08 0x0C
	_u32	mode;		// 0x10	- Mode setting	
	_u32	null1[3];	// 0x14 0x18 0x1C
	_u32	stat;		// 0x20 - Status	
	_u32	null2[7];	// 0x24 0x28 0x2C 0x30 0x34 0x38 0x3C
	_u32	tag0;		// 0x40	- GIFtag (bits 31-0) immediately before	
	_u32	null3[3];	// 0x44 0x48 0x4C
	_u32	tag1;		// 0x50	- GIFtag (bits 63-32) immediately before	
	_u32	null4[3];	// 0x54 0x58 0x5C
	_u32	tag2;		// 0x60	- GIFtag (bits 95-64) immediately before
	_u32	null5[3];	// 0x64 0x58 0x5C
	_u32	tag3;		// 0x70	- GIFtag (bits 127-96) immediately before	
	_u32	null6[3];	// 0x74 0x48 0x4C
	_u32	cnt;		// 0x80	- Transfer status counter	
	_u32	null7[3];	// 0x84 0x58 0x5C
	_u32	p3cnt;		// 0x90	- PATH3 transfer status counter	
	_u32	null8[3];	// 0x94 0x58 0x5C
	_u32	p3tag;		// 0xa0	- PATH3 tag value


} GIFRegisters;



/********************************************************************************************
*					GIF STRUCTURE															*		
*********************************************************************************************/

typedef struct _GIFGSInterface {

	GIFRegisters * Reg;

}GIFGSInterface;


extern GIFGSInterface GIF;





/********************************************************************************************
*					GIF FUNCTIONS DECLARATIONS												*		
*********************************************************************************************/

void 	GIF_init	( void );
void 	GIF_Log 	( char * ftext  , ...   );

void	GIF_Write 	( _u64 fPhysicalAddress, _u128 fVal, 	_u8 fDataType );
void	GIF_Read 	( _u64 fPhysicalAddress, _u128 * fVal, 	_u8 fDataType );


void GIF_PATH1_Transfert( _u32 *memAdd );
void GIF_PATH2_Transfert( _u32 *memAdd, _u32 size );
void GIF_PATH3_Transfert( _u32 *memAdd, _u32 size );


#endif








