/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: atreides.muaddib@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/
#include <stdlib.h>
#include <stdio.h>


#include "nSX2.h"
#include "PS2Edefs.h"


#include "IntrC.h"

 
/* Interrupt Controller*/
INTController 		IntC;


/* Bad Way .... need to clean */
char DMA_STDOUT[256];
_u16	DMA_STDOUT_count =  0;


 
/* LOG Function for the Interrupt Controller*/
void INTC_Log ( char * ftext  , ...   ) {
	char INTC_LogText[ FILE_NAME_LENGHT ];

	if ( debug_intc == 1 ) {

    	va_list ap;
		va_start (ap, ftext);
  		vsprintf (INTC_LogText, ftext, ap);
   		va_end (ap);

		sprintf(INTC_LogText,"%s\n",INTC_LogText);
		PrintToINTCOutput 		(  INTC_LogText ) ;
	}

}



/* Interrupt Controller Initialisation */
void	INTC_Init( void ) {
	_u16	boucle;

#ifdef EMU_LOG
	EMU_LOG("[INTC]Initialisation");
#endif
	/* Get pointer from Special Memory Area */
	IntC.Reg = ( InteruptControllerRegisters * ) &SpecialAreaMemory[   ( INTC_STAT & 0xFFFF )  ];

	/* At initialisation , no handler is active */
	
	for ( boucle = 0; boucle < BIOS_MAX_INTC_HANDLER; boucle++ ) {

		IntC.IntByCause[ boucle ].Numero_ID		= 0;
		IntC.IntByCause[ boucle ].Nb_handler	= 0;
		IntC.IntByCause[ boucle ].Status 		= 0 ;
		IntC.IntByCause[ boucle ].HandlerAdd	= NULL;
	}

	/* At initialisation , no handler is active */	


	for ( boucle = 0; boucle < BIOS_MAX_DMA_HANDLER; boucle++ ) {

		IntC.IntByDMA[ boucle ].Numero_ID	= 0;
		IntC.IntByDMA[ boucle ].Nb_handler	= 0;
		IntC.IntByDMA[ boucle ].Status 		= 0 ;
		IntC.IntByDMA[ boucle ].HandlerAdd	= NULL;
	}


	/* Interrupt Disable*/
	IntC.IntrEnable 		= FALSE;

	IntC.ExecutingHandler	= NO_EXECUTING_HANDLER;

	/* All Interrupts are Enable */
	IntC.Reg->mask = 0xFFFF;
	IntC.Reg->stat = 0x0000;

}


void	INTC_Shutdown ( void ) {
	int boucle;

	LinkedAddress * CurrentHandlerAdd;
	LinkedAddress * SavedHandlerAdd;

	for ( boucle = 0; boucle < BIOS_MAX_INTC_HANDLER; boucle++ ) {

		SavedHandlerAdd 	= IntC.IntByCause[ boucle ].HandlerAdd;

		while ( SavedHandlerAdd != NULL ) {
			CurrentHandlerAdd = SavedHandlerAdd->next;
			
			if ( SavedHandlerAdd != NULL ) free ( SavedHandlerAdd );
			
			SavedHandlerAdd = CurrentHandlerAdd;
		}
	}

	for ( boucle = 0; boucle < BIOS_MAX_DMA_HANDLER; boucle++ ) {

		SavedHandlerAdd 	= IntC.IntByDMA[ boucle ].HandlerAdd;

		while ( SavedHandlerAdd != NULL ) {
			CurrentHandlerAdd = SavedHandlerAdd->next;
			
			if ( SavedHandlerAdd != NULL ) free ( SavedHandlerAdd );
			
			SavedHandlerAdd = CurrentHandlerAdd;
		}
	}

#ifdef EMU_LOG
	EMU_LOG("[INTC]Shutdown");
#endif


}

/* Interrupt Controller Write Registers*/
void	INTC_Write ( _u64 fPhysicalAddress , _u128 fVal , _u8 fDataType ) {

	_u8		boucle;
	_u32	Mask;

	_u64 	RealAdd;

	if ( (fPhysicalAddress != INTC_STAT ) || (fPhysicalAddress != INTC_MASK ) ) {
	
		RealAdd 		= fPhysicalAddress & 0xFFFF;
		switch ( fDataType ) {

			case DATA_QUAD:	MipsMemoryWriteDW 	( SpecialAreaMemory, fVal.lo, RealAdd );
							MipsMemoryWriteDW 	( SpecialAreaMemory, fVal.hi, RealAdd + 8 );break;

				
			case DATA_DW:	MipsMemoryWriteDW	( SpecialAreaMemory, fVal.lo, RealAdd );break;
				
			case DATA_W:	MipsMemoryWriteW	( SpecialAreaMemory, fVal.lo, RealAdd );break;

			case DATA_HW:	MipsMemoryWriteHW	( SpecialAreaMemory, fVal.lo, RealAdd );break;

			case DATA_B:	MipsMemoryWriteB	( SpecialAreaMemory, fVal.lo, RealAdd );break;
										
			default:		

#ifdef EMU_LOG
	EMU_LOG("MemSpecial:Write DataType Not Handled\n");
#endif


							MipsCpu.halt = 1;break;
		}

	}
	
	switch ( fPhysicalAddress  ) {
		
		case INTC_STAT:	/* INT C Status */
#ifdef INTC_LOG
			INTC_LOG("%8.8lx:[INTC][WR]INTC_STAT : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
#endif
			
			IntC.Reg->stat = IntC.Reg->stat | fVal.lo;

			/* Check if Interrupt Raises */
			for ( boucle = 0 ; boucle < 11 ; boucle ++ ) INTC_RaiseIntByCause (  boucle );

		break;

 		
		case INTC_MASK:		/* INT C Mask */
#ifdef INTC_LOG
			INTC_LOG("%8.8lx:[INTC][WR]INTC_MASK : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
#endif
			
			//Reversed When 1 is written
			for ( boucle = 0 ;boucle < 16 ; boucle++ ) {
				Mask = 	( 1 << boucle );

				if ( (fVal.lo & Mask) == Mask ) {

					if ( ( IntC.Reg->mask & Mask ) == Mask ) {
						IntC.Reg->mask = IntC.Reg->mask & ~( Mask );
					}else {
						IntC.Reg->mask = IntC.Reg->mask | ( Mask );
					}
				}
			}

		break;


		/* Not the right Place , here */
		case STD_OUT_DATA:  	
#ifdef INTC_LOG
			INTC_LOG("%8.8lx:[INTC][WR]STD_OUT_DATA : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
#endif

			if ( DMA_STDOUT_count == 0 ) memset ( DMA_STDOUT,0,256);

			DMA_STDOUT[DMA_STDOUT_count] = (_u8)fVal.lo;

			if ( DMA_STDOUT[DMA_STDOUT_count] == 0xa ) {
				DMA_STDOUT[DMA_STDOUT_count + 1 ] = 0;
#ifdef EMU_LOG
				EMU_LOG("[STD_OUT] %s\n",DMA_STDOUT);
#endif
				DMA_STDOUT_count = 0;
			}

			DMA_STDOUT_count++;			

		break;
		
		default: {
#ifdef EMU_LOG
			EMU_LOG("%8.8lx:[INTC][WR]???? : 0x%8.8lx ", (long)MipsCpu.PC,fPhysicalAddress);
#endif

		break;
		}
		
	}






}


/* Interrupt Controller Read Registers*/
void	INTC_Read ( _u64 fPhysicalAddress , _u128 * fVal , _u8 fDataType ) {

	_u64 	RealAdd;

#ifdef INTC_LOG		
	switch ( fPhysicalAddress  ) {
	
		case INTC_STAT:		//
			//INTC_LOG("%8.8lx:[INTC][RD]INTC_STAT : 0x%8.8lx ", (long)MipsCpu.PC, IntC.Reg->stat);
		break;

		case INTC_MASK:  	//
			INTC_LOG("%8.8lx:[INTC][RD]INTC_MASK : 0x%8.8lx ", (long)MipsCpu.PC, IntC.Reg->mask);
		break;

		case STD_OUT_STATUS: //
			INTC_LOG("%8.8lx:[INTC][RD]STD_OUT_STATUS : 0x%8.8lx ", (long)MipsCpu.PC, IntC.Reg->mask);
		break;
		
		default: {
			EMU_LOG("%8.8lx:[INTC][RD]???? : 0x%8.8lx ", (long)MipsCpu.PC,fPhysicalAddress);
		break;
		}
		
	}
#endif
 
	RealAdd 		= fPhysicalAddress & 0xFFFF;

	switch ( fDataType ) {

		case DATA_QUAD:	MipsMemoryReadDQ 	( SpecialAreaMemory, fVal, RealAdd );break;
					
		case DATA_DW:	MipsMemoryReadDW	( SpecialAreaMemory, fVal, RealAdd );break;

	    case DATA_HW:	MipsMemoryReadHW	( SpecialAreaMemory, fVal, RealAdd );break;
		
		case DATA_W:	MipsMemoryReadW 	( SpecialAreaMemory, fVal, RealAdd );break;

		case DATA_B:	MipsMemoryReadB		( SpecialAreaMemory, fVal, RealAdd );break;

		default: 		

#ifdef EMU_LOG
	EMU_LOG("MemSpecial:Read DataType Not Handled\n");
#endif

						MipsCpu.halt = 1;break;
	}

}






