/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: atreides.muaddib@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/
#include <stdlib.h>
#include <stdio.h>

#include <math.h>
#include <string.h>


#include "nSX2.h"
#include "PS2Edefs.h"

#include "VU.h"
#include "VU1_micromode_tables.h"

#include "VIF.h"


/********************************************************************************************/
/*																	*/
/********************************************************************************************/




/*Limit function*/
int _limit ( int a , int max ) { return ( a > max ? max : a );}

/*
Compute the Unpack Data Size . unit is 128 bits
This isn't magic function , everything is in the docs
*/
_u32	VIF_ComputePackLenght ( _u8 fVIF, _u8 vn , _u8 vl , _u16 fsize) {
	_u32 wl;
	_u32 cl;

	int	temp32;
	int	returnValue;

	/* Extract wl and cl */
	wl = ( (_u32)(VIF[ fVIF ].Reg->cycle) >> 8 ) & 0xFF;
	cl = (_u32)(VIF[ fVIF ].Reg->cycle) & 0xFF;

	/* Compute the size of the Packet */
	if (wl <= cl) {
			returnValue = ((int)ceil((double)(((32>>vl) * (vn+1)) * fsize)/32));
	} else {
		temp32 = cl * (fsize / wl) + _limit(fsize%wl, cl);
		returnValue = ((int)ceil((double)(((32>>vl) * (vn+1)) * temp32)/32));
	}
	
	/* Return the Size */
	return returnValue;
}





/*
Copy Program From EE Memory to VU Memry
*/
void VIF_CopyMicroProgram ( _u8 fVIF, _u32 * CurrentData , _u32 fsize ) {
	_u32 * dataSrc;				//pointer to data src ( Mips Memory )
	_u32 * dataDest;			//pointer to data dest ( VectorUnit Memory )

	_u32	dataDestAdd;		//Addr of Destination Data in the VectorUnit1 Memory

	
	/* Extract the Destination Addr form the VIF Command*/
	dataDestAdd = ((* CurrentData) & 0x3FF ) * 8;

	/* The first data is just after the command */
	dataSrc = CurrentData + 1;	

	/* Compute the destination pointer */
	switch ( fVIF ) {
		case VIF0: dataDest = (_u32*)(VectorUnit0.MicroMem + dataDestAdd);break;
		case VIF1: dataDest = (_u32*)(VectorUnit1.MicroMem + dataDestAdd);break;
	}

  

	/* Print debug info */
#ifdef VIF_LOG_TR
		VIF_LOG_TR("[VIF%1.1d][TR]VU.micro add : 0x%8.8x , size : %d (_u64) = %d ( vuOP ) Copy : %d (_u8) ", 
			fVIF,dataDestAdd,fsize,fsize*2,sizeof(_u32) * ( fsize * 2 ));
#endif

	/* Copy the PRG ( fsize is 64bit unit ) */ 
	memcpy ( dataDest, dataSrc, sizeof(_u32) * ( fsize * 2 ));

	/* Write the Copied MicroProgram to the output file */
	VIF1_DumpMicroCodetoLogFile( dataDest , ( fsize * 2 ) );

}


/*
Initialisation before Activation of the MicroProgram
MUST BE DONE WHITH VIF_CMD_MSCAL , VIF_CMD_MSCNT, VIF_CMD_MSCALF
*/
void VIF_InitMicroProgramActivation ( _u8 fVIF ) {
	_u128 temp128;

	/* Transfer the TOPS reg to TOP reg ( access with XTOP )*/
	VIF[ fVIF ].Reg->top 	= 	(_u32)VIF[ fVIF ].Reg->tops;

	/* Transfer the ITOPS reg to ITOP reg ( access with XITOP )*/
	VIF[ fVIF ].Reg->itop 	= 	(_u32)VIF[ fVIF ].Reg->itops;
}


