/*  Pcsx2 - Pc Ps2 Emulator
 *  Copyright (C) 2002-2003  Pcsx2 Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <fcntl.h>
#ifdef __WIN32__
#include <io.h>
#else
#include <unistd.h>
#endif
#include <string.h>
#include <sys/stat.h>
#include <errno.h>

#include "Rpc_services.h"
#include "../CDVDisodrv.h"
#include "../SJdata.h"

//should be 0x40000000; not 0x40 :(
#define CUSTOMFS			0x40			//ps2lib workaround
#define CDFILE				(CUSTOMFS>>0)	//0x40
#define SJFILE				(CUSTOMFS>>1)	//0x20
//that makes 32 files per filesystem :(
//i hope that Sjeep will recompile PGEN soon,
//with a non-buggy version of ps2lib

int *addr=NULL, *addr2;
struct unk{
	int		sema;
	int		fcn_index;
	void	*addr;
	int		size;
	int		result;
	int		a,
			b,
			c;
} unk;
///////////////////////////////////////////////////
//FILEIO_OPEN	0
//int sceOpen(char *filename, int flag);
//SCE_RDONLY      0x0001
//SCE_WRONLY      0x0002
//SCE_RDWR        0x0003
//SCE_NBLOCK      0x0010  /* Non-Blocking I/O */
//SCE_APPEND      0x0100  /* append (writes guaranteed at the end) */
//SCE_CREAT       0x0200  /* open with file create */
//SCE_TRUNC       0x0400  /* open with truncation */
//SCE_NOBUF       0x4000  /* no device buffer and console interrupt */
//SCE_NOWAIT      0x8000  /* asyncronous i/o */

//SCE_USING       0x10000000/* io descript usage bit */
//SCE_NOWBDC      0x20000000/* not write back d cashe */
void fileio_open(struct t_sif_rpc_call*	call, u32 *bin, u32 *bout){
	int		flag=0, inflag;
	char	*name, *device, *inname;
	char	file2open[256];
	int		sema=-1;




	if (call->send_size==sizeof(struct fileio_OPENn)//fileio v2.08 and up
						+sizeof(struct unknown)+4){	//new style
		inname=((struct fileio_OPENn*)(bin+3))->name;
		inflag=((struct fileio_OPENn*)(bin+3))->flag;
		sema=((struct unknown*)bin)->sema;
	}else if (strncmp(RPCVER, "2100", 4)>0){//fileio v2.07, only
		inname=((struct fileio_OPENn*)(bin))->name;
		inflag=((struct fileio_OPENn*)(bin))->flag;
	}else{												//old style
		inname=((struct fileio_OPEN*)bin)->name;
		inflag=((struct fileio_OPEN*)bin)->flag;
	}

#ifdef RPC_LOG
	RPC_LOG("[RPC:fileio] open(\"%s\", mode=%08X)\n", inname, inflag);
#endif

	if (inflag & 0xFFFFC010){//other flags are not supported for now
#ifdef RPC_LOG
		RPC_LOG("             unsuported flags: 0x%08X\n", inflag & 0xFFFFC010);
#endif
		*bout=-1;
		return;
	}
	if (inflag & 0x0001)		flag = O_RDONLY;
	if (inflag & 0x0002)		flag = O_WRONLY;
	if ((inflag & 0x0003)==3)	flag = O_RDWR;
	if (inflag & 0x0100)		flag|= O_APPEND;
	if (inflag & 0x0200)		flag|= O_CREAT;
	if (inflag & 0x0400)		flag|= O_TRUNC;
#ifdef __WIN32__
								flag|= O_BINARY;
#endif

	name=device=inname; while(*name && *name!=':') name++; if (*name==':') {*name=0; name++;}
	if ((strncmp(device, "cdrom", strlen("cdrom"))==0)||
		(strncmp(device, "cdfs", strlen("cdfs"))==0)){
		inflag=1;//workaround for those who opens files from cd without RDONLY
		*bout=CDVDFS_open(name, inflag);
		if (*bout>=0) *bout|=CDFILE;		//marks as cdrom file
	}else
	if ((strncmp(device, "sjdata", strlen("sjdata"))==0)){
		inflag=1;//workaround for those who opens files from cd without RDONLY
		*bout=SjDATA_open(name, inflag);
		if (*bout>=0) *bout|=SJFILE;		//marks as SjDATA file
	}else
	if ((strncmp(device, "rom", strlen("rom"))) &&
		(strncmp(device, "mc", strlen("mc"))) &&
		(strncmp(device, "host", strlen("host")))){
#ifdef RPC_LOG
		RPC_LOG("             unsuported device: \"%s\"\n", device);
#endif
		*bout=-1;
		return;
	}else{
		if ((*name!='\\') && 
			(*name!='/') && 
			(strncmp(name+1, ":\\", 2))){
#ifdef	__LINUX__
			sprintf(file2open, "host/%s", name);
#else
//			RPC_LOG("device:%s , name:%s",device,name);

			sprintf(file2open, "%s\\%s\\%s", muadCurDir,device,name);
#endif
		}else {
//			sprintf(file2open, "%s%s", ((*name=='/')||(*name=='\\')?"d:":""), name);
//			strcpy(file2open, name);
			sprintf(file2open, "%s\\%s%s", muadCurDir,device, name);
//			RPC_LOG("2");
		}


#ifdef RPC_LOG		
		RPC_LOG("File_to_open : %s",file2open);
#endif
		*bout = open(file2open, flag, S_IREAD | S_IWRITE);
#ifdef RPC_LOG
		if (*bout==-1)
			switch(errno){
			case ENOENT:	RPC_LOG("             ENOENT File or path not found\n");break;
			case EMFILE:	RPC_LOG("             EMFILE No more file handles available (too many open files)\n");break;
			case EINVAL:	RPC_LOG("             EINVAL Invalid oflag or pmode argument\n");break;
			case EEXIST:	RPC_LOG("             EEXIST _O_CREAT and _O_EXCL flags specified, but filename already exists\n");break;
			case EACCES:	RPC_LOG("             EACCES Tried to open read-only file for writing, or files sharing mode does not allow specified operations, or given path is directory\n");break;
			default:		RPC_LOG("             errno=%d\n", errno);break;
		}
#endif
	}
	
#ifdef RPC_LOG
	if (*(int*)bout < 0){
		RPC_LOG("             [%s] %s => ERROR: %d\n", device, name, *bout);
	} else {
		RPC_LOG("             [%s] %s => fd=0x%08X\n", device, name, *bout);
	}
#endif
	if ((*(int*)bout >= 0) && (sema>=0)){
#if _SIF_PLUGIN_
		g_SifInitialize.pSignalSema(sema);
#else
		bSema[sema].sema.count++;
#endif

		unk.sema=sema;
		unk.fcn_index=FILEIO_OPEN;
		unk.addr=((struct unknown*)bin)->addr;
		unk.size=((struct unknown*)bin)->size;
		unk.result=*bout;
		memcpy(PSM((int)addr), &unk, sizeof(struct unk));

		//this is job of 0x80000011 cmd handler
		(int)unk.addr&=0x1FFFFFFF;//hey linuzappz; what do ya say about this?
		memcpy(PSM((int)unk.addr), &unk.result, unk.size);
	}
}

///////////////////////////////////////////////////
//FILEIO_CLOSE	1
//int sceClose(int fd);		0=OK	-1=fail
void fileio_close(struct t_sif_rpc_call*	call, u32 *bin, u32 *bout){
	register struct fileio_CLOSE	*in=(struct fileio_CLOSE *)bin;
	int		sema=-1;

	if (call->send_size==sizeof(struct fileio_CLOSE)//fileio v2.08 and up
		+sizeof(struct unknown)+4){
		in=(struct fileio_CLOSE *)(((char*)in)+sizeof(struct unknown));
		sema=((struct unknown*)bin)->sema;
	}
#ifdef RPC_LOG
	RPC_LOG("[RPC:fileio] close(fd=0x%08X)\n", in->fd);
#endif
	if (in->fd & CDFILE)	*bout = CDVDFS_close(in->fd & (~CDFILE));	else
	if (in->fd & SJFILE)	*bout = SjDATA_close(in->fd & (~SJFILE));	else
							*bout = ((in->fd>2) ? close(in->fd) : -1);
	if (sema>=0){
#if _SIF_PLUGIN_
		g_SifInitialize.pSignalSema(sema);
#else
		bSema[sema].sema.count++;
#endif

		unk.sema=sema;
		unk.fcn_index=FILEIO_CLOSE;
		unk.addr=((struct unknown*)bin)->addr;
		unk.size=((struct unknown*)bin)->size;
		unk.result=*bout;
		memcpy(PSM((int)addr), &unk, sizeof(struct unk));

		//this is job of 0x80000011 cmd handler
		(int)unk.addr&=0x1FFFFFFF;
		memcpy(PSM((int)unk.addr), &unk.result, unk.size);
	}
}

///////////////////////////////////////////////////
//FILEIO_READ	2
//int sceRead(int fd, void *buffer, int size);
void fileio_read(struct t_sif_rpc_call*	call, u32 *bin, u32 *bout){
	register struct fileio_READ	*in=(struct fileio_READ *)bin;
	int		sema=-1;

	if (call->send_size==sizeof(struct fileio_READ)//fileio v2.08 and up
		+sizeof(struct unknown)+4){
		in=(struct fileio_READ *)(((char*)in)+sizeof(struct unknown));
		sema=((struct unknown*)bin)->sema;
	}
#ifdef RPC_LOG
	RPC_LOG("[RPC:fileio] read(fd=0x%08X, buf=0x%08X, size=%d)(_intr_data=0x%08X)\n", in->fd, in->buffer, in->size, in->_intr_data);
#endif
	(u32)in->buffer &= 0x7FFFFFFF;
	if (in->fd & CDFILE)	*bout = CDVDFS_read(in->fd & (~CDFILE), PSM((int)in->buffer), in->size);else
	if (in->fd & SJFILE)	*bout = SjDATA_read(in->fd & (~SJFILE), PSM((int)in->buffer), in->size);else
							*bout = read(in->fd, PSM((int)in->buffer), in->size);
	if (*bout)
		if (sema==-1){
//  <--------------size------------->
//  /-------------------------------/
//  | <-ssize-> | asize | <-esize-> |
//  \-------------------------------/
//  ^in->buffer ^abuf   ^abuf+asize
			char *abuf=(char*)(((int)in->buffer+15) & 0xFFFFFFF0);
			int ssize=abuf-(char*)in->buffer, esize=(in->size-ssize) % 16, asize=in->size-ssize-esize;
			struct _rests rests={{ssize, esize, in->buffer, abuf+asize}};
			memcpy(rests.sbuffer, PSM((int)rests.rest.sbuf), ssize);
			memcpy(rests.ebuffer, PSM((int)rests.rest.ebuf), esize);
			memcpy(PSM((int)in->_intr_data), &rests, sizeof(struct _rests));
#ifdef RPC_LOG
			if (ssize || esize)
			RPC_LOG("                  startsize=%d|0x%08X<-%d->0x%08X|endsize=%d\n",
										ssize, abuf, asize, abuf+asize, esize);
#endif
		}else{
			struct fileio_READr r;
			char *abuf=(char*)(((int)in->buffer+15) & 0xFFFFFFF0);
#if _SIF_PLUGIN_
		g_SifInitialize.pSignalSema(sema);
#else
		bSema[sema].sema.count++;
#endif

			r.sema=sema;
			r.fcn_index=FILEIO_READ;
			r.addr=((struct unknown*)bin)->addr;
			r.size=((struct unknown*)bin)->size;
			r.result=*bout;
			r.ssize=abuf-(char*)in->buffer;
			r.esize=(in->size-r.ssize) % 16;
			r.sbuf=in->buffer;
			r.ebuf=abuf+in->size-r.ssize-r.esize;
			memcpy(r.sbuffer, PSM((int)r.sbuf), r.ssize);
			memcpy(r.ebuffer, PSM((int)r.ebuf), r.esize);
			memcpy(PSM((int)addr), &r, sizeof(struct fileio_READr));

			//this is job of 0x80000011 cmd handler
			(int)r.addr&=0x1FFFFFFF;
			memcpy(PSM((int)r.addr), &r.result, r.size);//should be sizeof(struct fileio_READr)-0x10?
		}
}

///////////////////////////////////////////////////[NOT FULLY COMPATIBLE]
//FILEIO_WRITE	3
//int sceWrite(int fd, void *buffer, int size);
void fileio_write(struct t_sif_rpc_call*	call, u32 *bin, u32 *bout){
	register struct fileio_WRITE *in=(struct fileio_WRITE *)bin;
#ifdef RPC_LOG
	RPC_LOG("[RPC:fileio] write(fd=0x%08X, buf=0x%08X, size=%d)(mis=0x%08X)\n", in->fd, in->buffer, in->size, in->mis);
#endif
	if (in->fd & CDFILE)*bout = CDVDFS_write(in->fd & (~CDFILE), PSM((int)in->buffer), in->size);else
	if (in->fd & SJFILE)*bout = SjDATA_write(in->fd & (~SJFILE), PSM((int)in->buffer), in->size);
	else
		if  (in->fd==0)
			*bout=-1;		else
		if ((in->fd==1)||
			(in->fd==2)){
			SysPrintf("%s\n", PSM((int)in->buffer));//bonus: \n :D
			*bout = in->size;
		}					else
			*bout = write(in->fd, PSM((int)in->buffer), in->size);
}

///////////////////////////////////////////////////
//FILEIO_LSEEK	4
//int sceLseek(int fd, int offset, int where);
void fileio_lseek(struct t_sif_rpc_call*	call, u32 *bin, u32 *bout){
	register struct fileio_LSEEK *in=(struct fileio_LSEEK *)bin;
	int		sema=-1;

	if (call->send_size==sizeof(struct fileio_LSEEK)//fileio v2.08 and up
		+sizeof(struct unknown)+4){
		in=(struct fileio_LSEEK *)(((char*)in)+sizeof(struct unknown));
		sema=((struct unknown*)bin)->sema;
	}
#ifdef RPC_LOG
	RPC_LOG("[RPC:fileio] lseek(fd=0x%08X, off=%d, %s)", in->fd, in->offset, in->ref==0 ? "SEEK_SET": in->ref==1 ? "SEEK_CUR":"SEEK_END");
#endif
	if (in->fd & CDFILE)	*bout = CDVDFS_lseek(in->fd & (~CDFILE), in->offset, in->ref);else
	if (in->fd & SJFILE)	*bout = SjDATA_lseek(in->fd & (~SJFILE), in->offset, in->ref);else
							*bout = lseek(in->fd, in->offset, in->ref);
#ifdef RPC_LOG
	RPC_LOG(" => %d(0x%08X)\n", *bout, *bout);
#endif
	if (sema>=0){
#if _SIF_PLUGIN_
		g_SifInitialize.pSignalSema(sema);
#else
		bSema[sema].sema.count++;
#endif

		unk.sema=sema;
		unk.fcn_index=FILEIO_LSEEK;
		unk.addr=((struct unknown*)bin)->addr;
		unk.size=((struct unknown*)bin)->size;
		unk.result=*bout;
		memcpy(PSM((int)addr), &unk, sizeof(struct unk));

		//this is job of 0x80000011 cmd handler
		(int)unk.addr&=0x1FFFFFFF;
		memcpy(PSM((int)unk.addr), &unk.result, unk.size);
		*bout=(*bout>=0);//for all functions with sema, *bout should contain 0 for failure or smth. else for success
	}
}

///////////////////////////////////////////////////
//FILEIO_FSINIT	0xFF
//char* sceFsInit(void *addr);
void fileio_fsinit(struct t_sif_rpc_call*	call, u32 *bin, u32 *bout){
	if (call->send_size==4)		addr=addr2=(int*)bin[0];	else
	if (call->send_size==8){	addr=(int*)bin[0];addr2=(int*)bin[1];}
#ifdef RPC_LOG
	RPC_LOG("[RPC:fileio] sceFsInit(addr=0x%08X, addr2=0x%08X) => %4s\n", addr, addr2, RPCVER);
#endif
	*bout=*(int*)RPCVER;
}

//ENTRYPOINT
void sifcall_fileio(struct t_sif_rpc_call*	call, u32 *bin, u32 *bout){
	switch(call->rpc_number){
	case FILEIO_OPEN:	fileio_open(call, bin, bout);	break;
	case FILEIO_CLOSE:	fileio_close(call, bin, bout);	break;
	case FILEIO_READ:	fileio_read(call, bin, bout);	break;
	case FILEIO_WRITE:	fileio_write(call, bin, bout);	break;
	case FILEIO_LSEEK:	fileio_lseek(call, bin, bout);	break;
	case FILEIO_FSINIT:	fileio_fsinit(call, bin, bout);	break;
	default:
#ifdef RPC_LOG
		RPC_LOG("[RPC:fileio] <0x%02X> function is not implemented (0x%04X)\n", call->rpc_number, FILEIO_VER);
#endif
		*bout=-1;	//general fail
	}
}