/*  Pcsx2 - Pc Ps2 Emulator
 *  Copyright (C) 2002-2003  Pcsx2 Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __FILEIO_H__
#define __FILEIO_H__

#define FILEIO_VER	0x101

//////////////////////////////////////////////// 80000001 fileio rpc service
#define FILEIO_OPEN	0
#define FILEIO_CLOSE	1
#define FILEIO_READ	2
#define FILEIO_WRITE	3
#define FILEIO_LSEEK	4
#define FILEIO_IOCTL	5
#define FILEIO_REMOVE	6
#define FILEIO_MKDIR	7
#define FILEIO_RMDIR	8
#define FILEIO_DOPEN	9
#define FILEIO_DCLOSE	10
#define FILEIO_DREAD	11
#define FILEIO_GETSTAT	12
#define FILEIO_CHSTAT	13
#define FILEIO_FORMAT	14
#define FILEIO_ADDDRV	15
#define FILEIO_DELDRV	16
//*************************** not in version 0x101
#define FILEIO_RENAME	17
#define FILEIO_CHDIR	18
#define FILEIO_SYNC		19
#define FILEIO_MOUNT	20
#define FILEIO_UMOUNT	21
#define FILEIO_LSEEK64	22
#define FILEIO_DEVCTL	23
#define FILEIO_SYMLINK	24
#define FILEIO_READLINK	25
#define FILEIO_IOCTL2	26
#define FILEIO_FSINIT	255
//****************************/

struct sce_stat {
	unsigned int st_mode;/* File mode
				bit 0 Execute permission (other)
				bit 1 Write permission (other)
				bit 2 Read permission (other)
				bit 3 Execute permission (group)
				bit 4 Write permission (group)
				bit 5 Read permission (group)
				bit 6 Execute permission (user)
				bit 7 Write permission (user)
				bit 8 Read permission (user)
				bit 9 Reserved
				bit10 Reserved
				bit11 Reserved
				bit12-15 File type
					1 Directory
					2 Normal file
					4 Symbolic link */
	unsigned int st_attr;// Flag compatible with memory card mode
	unsigned int st_size;
	unsigned char st_ctime[8];// Creation time
	unsigned char st_atime[8];/* This field is updated at the same time as last access time
				and last update time. */
	unsigned char st_mtime[8];/* Last update time
				byte0 Reserved
				byte1 Seconds
				byte2 Minutes
				byte3 Hours
				byte4 Day
				byte5 Month
				byte6-7 (4 digits) */
	unsigned int st_hisize;// File size (64 bit)
	unsigned int st_private[6];/*	word0 uid
					word1 gid
					word2 Number of zones used by the file */
};

struct unknown{
	int		sema;	//00
	void	*addr;	//04
	int		size;	//08
};

struct fileio_DRIVER{	//struct fileio_driver from ps2lib
	char	*driver;
	int	unk1;
	int	version;
	char	*description;
	void	**func_list;
};

//////////////////////////////////////////////////////// functions params
struct fileio_OPEN{			//0
	int		flag;		//000
	char		name[1024];	//004
};					//404

struct fileio_OPENn{			//0	//new version
	int		flag;		//000
	int		mode;		//004
	char		name[1024];	//008
};					//408

struct fileio_CLOSE{			//1
	int		fd;		//000
};					//004

struct fileio_READ{			//2
	int	fd;			//000
	void	*buffer;		//004
	int	size;			//008
	void	*_intr_data;		//00C
};					//010

struct fileio_READr{			//2	//retrieve
	int	sema,			//000
		fcn_index;		//004
	void	*addr;			//008
	int	size,			//00C
		result,			//010
		ssize,			//014
		esize;			//018
	void	*sbuf,			//01C
		*ebuf;			//020
	char	sbuffer[64],		//024
		ebuffer[64];		//064
};					//0A4

struct fileio_WRITE{			//3
	int	fd;			//000
	void	*buffer;		//004
	int	size;			//008
	int	mis;			//00C
};					//010

struct fileio_WRITEn{			//3
	int	fd;			//000
	void	*buffer;		//004
	int	size;			//008
	int	mis;			//00C
	char	misbuffer[16];		//010	NOT SURE OF SIZE; just a guess:P
};					//020	check send_size and tell me;)

struct fileio_LSEEK{			//4
	int	fd;			//000
	int	offset;	//pos		//004
	int	ref;	//mode		//008
};					//00C

struct fileio_IOCTL{			//5
	int	fd;			//000
	int	io_code;		//004	NOT VERY SURE ABOUT
};					//008

struct fileio_IOCTLn{			//5
	int	fd;			//000
	int	io_code;		//004
	char	buffer[1024];		//008
	int	_unk414;		//408
	int	_unk418;		//40C
	int	use_buffer;		//410
};					//414

struct fileio_REMOVE{			//6
	char	name[1024];		//000
};					//400

struct fileio_MKDIR{			//7
	char		name[1024];	//000
};					//400

struct fileio_MKDIRn{			//7
	int		mode;		//000
	char		name[1024];	//004
};					//404

struct fileio_RMDIR{			//8
	char	name[1024];		//000
};					//400

struct fileio_DOPEN{			//9
	char	name[1024];		//000
};					//400

struct fileio_DCLOSE{			//A
	char	fd;			//000
};					//004

struct fileio_DREAD{			//B
	char	fd;			//000
	void	*buf;			//004
};					//008

struct fileio_GETSTAT{			//C
	struct sce_stat	*buf;		//000
	char		name[1024];	//004
};					//404

struct fileio_CHSTAT{			//D
	int		cbit;		//000
	struct sce_stat	buf;		//004
	char		name[1024];	//044
};					//444

struct fileio_FORMAT{			//E
	char	devname[1024];		//000
	char	blockdevname[1024];	//400
	char	arg[1024];		//800
	int	arglen;			//C00
};					//C04

struct fileio_ADDDRV{			//F
	struct fileio_DRIVER	*driver;//000
};					//004

struct fileio_DELDRV{			//F
	char	name[1024];		//000
};					//400

//////////////////////////////////////////////// 80000003 iopheap rpc service
#define IOPHEAP_ALLOC		1
#define IOPHEAP_FREE		2
#define IOPHEAP_LOAD		3
// *************************** not in version 0x101
#define IOPHEAP_ALLOCSYSMEMORY	4
// ***************************

struct fileio_LOADIOPHEAP{
	char *buffer;
	char name[0];
};

#endif//__FILEIO_H__