/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: atreides.muaddib@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/

#ifndef _BIOS_H
#define _BIOS_H

#include "nSX2_type.h"

/*	SEMAPHORE	*/

#define		BIOS_MAX_THREADS			256
#define		BIOS_THREAD_ROOMFREE		0
#define		BIOS_THREAD_ROOMBUSY		1


/*
Kernel Semaphore Structure
*/
typedef struct _SemaParam {
    _s32 currentCount;               //Current semaphore count
    _s32 maxCount;                   //Maximum semaphore count
    _s32 initCount;                  //Initial semaphore count
    _s32 numWaitThreads;             //Number of threads waiting on semaphore
    _s32 attr;                       //Semaphore attribute
    _u32 option;                   	//
} SemaParam;



typedef struct _Bios_Sema {

	SemaParam	* sema;
	_s16	WaitQueue[BIOS_MAX_THREADS];
   
} Bios_Sema;





/*
Threads
*/
#define	THREAD_ERROR		(-1)

#define	THS_RUN 			0x01 	//RUN state
#define THS_READY 			0x02 	//READY state
#define THS_WAIT 			0x04 	//WAIT state
#define THS_SUSPEND			0x08 	//SUSPEND state
#define	THS_WAITSUSPEND 	0x0c 	//WAIT-SUSPEND state
#define	THS_DORMANT 		0x10 	//DORMANT state

#define	THS_DESTROY 		0xEF 	// state

#define WAIT_TYPE_NO		0
#define WAIT_TYPE_REQUEST	1
#define WAIT_TYPE_SEMA		2

/* Values used to indicate WAIT type */
#define THS_WAITTYPE_NO 			0 		//Not in WAIT state
#define THS_WAITTYPE_WAKUPREQUEST  	1 		//Waiting for WAKUP request
#define	THS_WAITTYPE_SEMAPHORE		2 		//Waiting for semaphore


#define	DEFAULT_STACK_SIZE			4096

typedef struct _PS2_Thread {
	_s32 status; 				//Thread status
	_u32 entry_add;				//void (*entry)(void *); Entry address for execution 	(*)
	_u32 stack_add;				//void *stack; Stack address used by thread 			(*) (16-byte alignment)
	_s32 stackSize; 			//Stack size (in bytes: multiple of 16) 				(*)
	_u32 gpReg_add;				//void *gpReg; Value of GP (Global Pointer) register 	(*)
	_s32 initPriority;			//Initial priority (0 - 128) 							(*)
	_s32 currentPriority;		//Current priority (0 - 128)
	_u32 attr; 					//System Reserved (value is arbitrary)
	_u32 option; 				//Optional user-defined information
	_s32 waitType; 				//WAIT type
	_s32 waitId;				//semaphore ID if WAIT type is "semaphore"
	_s32 wakeupCount; 			//Wakeup request count

	
	
}PS2_Thread; 


typedef struct _LLThread {
	_u16	ID;
	_u8		RequestedStatus;

	PS2_Thread * thread;

	_u8		*Stack;

	_u32	EndOfHeap;
	_u32	HeapSize;

	_u128   GPR[34];
	_u128	FPR1[32];

	_u32	PC;

	struct _LLThread * previous;
	struct _LLThread * next;

	struct _LLThread * previousALL;
	struct _LLThread * nextALL;


}LLThread;



typedef	struct 	_Bios_Scheduler {


	_u16	TotalThread;
	_s16	RunningIDThread;

	LLThread * All_Threads;

	LLThread * Running_Thread;
	LLThread * Ready_Threads;	
	LLThread * Other_Threads;	

	_u16	MaxID;
	
} Bios_Scheduler;

void	THREADS_Init();
void	THREADS_Shutdown();

_u8		Thread_GetFreeRoomByPriority ( _u8 fPriority );
_u16 	Thread_GetFreeID ( void );

void 	InitQueue ( _s16 * fQueue );
void	AddLastQueue(_s16 * fQueue, _u16 fID);
void	AddFirstQueue( _s16 * fQueue, _u16 fID );
void	RemoveLastQueue( _s16 * fQueue );
void	RemoveFirstQueue( _s16 * fQueue );
void	RotateLeftQueue( _s16 * fQueue );
void	RemoveByIDQueue( _s16 * fQueue , _s16 fID);
_u8		IsPresentInQueue( _s16 * fQueue , _s16 fID) ;
void 	printQueue ( _s16 * fQueue );
void	UpdatePS2Scheduler ( void );

LLThread*	THREADS_Create( PS2_Thread * fPS2_Thread,	_u32 fState ) ;
void		THREADS_MoveToNotReady ( _u32	fID );
void		THREADS_MoveToReady ( _u32	fID );
void		THREADS_RotateReadyThread( void ) ;

void	THREADS_DeleteThreads ( _u32 fID );
void	THREADS_PutHighPriorityThreadFirst( void ) ;



#endif
