/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: atreides.muaddib@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/
#include <stdio.h>
#include <stdlib.h>

#include <strings.h>
#include <float.h>
#include <math.h>


#include "nSX2.h"

#include "r5900_main.h"
 
#include "bios_syscalls.h"
#include "bios_syscalls_table.h"
 
#include "PS2Edefs.h"

#include "Sif.h"

/* I must declare a function to avoid this kind of include */
#include "DmaC.h"

/* must create function to avoid this kind of declaration */
#include "IntrC.h"


#include "GS.h"




#define DECI2_PROTO_ANY		0x0000	/* Any */
#define DECI2_PROTO_DCMP	0x0001	/* Control Message */
#define DECI2_PROTO_I0TTYP	0x0110	/* IOP TTY 0 */
#define DECI2_PROTO_I1TTYP	0x0111	/* IOP TTY 1 */
#define DECI2_PROTO_I2TTYP	0x0112	/* IOP TTY 2 */
#define DECI2_PROTO_I3TTYP	0x0113	/* IOP TTY 3 */
#define DECI2_PROTO_I4TTYP	0x0114	/* IOP TTY 4 */
#define DECI2_PROTO_I5TTYP	0x0115	/* IOP TTY 5 */
#define DECI2_PROTO_I6TTYP	0x0116	/* IOP TTY 6 */
#define DECI2_PROTO_I7TTYP	0x0117	/* IOP TTY 7 */
#define DECI2_PROTO_I8TTYP	0x0118	/* IOP TTY 8 */
#define DECI2_PROTO_I9TTYP	0x0119	/* IOP TTY 9 */
#define DECI2_PROTO_IKTTYP	0x011f	/* IOP Kernel TTY */
#define DECI2_PROTO_DRFP0	0x0120	/* IOP File (host0:) for EE */
#define DECI2_PROTO_DRFP1	0x0121	/* IOP File (host1:) for IOP */
#define DECI2_PROTO_DRFP2	0x0122	/* IOP File (host2:) for CONF */
#define DECI2_PROTO_ISDBGP	0x0130	/* IOP System Debugging */
#define DECI2_PROTO_ITDBGP	0x0140	/* IOP Thread Debugging */
#define DECI2_PROTO_ILOADP	0x0150	/* IOP Load */
#define DECI2_PROTO_E0TTYP	0x0210	/* EE  TTY 0 */
#define DECI2_PROTO_E1TTYP	0x0211	/* EE  TTY 1 */
#define DECI2_PROTO_E2TTYP	0x0212	/* EE  TTY 2 */
#define DECI2_PROTO_E3TTYP	0x0213	/* EE  TTY 3 */
#define DECI2_PROTO_E4TTYP	0x0214	/* EE  TTY 4 */
#define DECI2_PROTO_E5TTYP	0x0215	/* EE  TTY 5 */
#define DECI2_PROTO_E6TTYP	0x0216	/* EE  TTY 6 */
#define DECI2_PROTO_E7TTYP	0x0217	/* EE  TTY 7 */
#define DECI2_PROTO_E8TTYP	0x0218	/* EE  TTY 8 */
#define DECI2_PROTO_E9TTYP	0x0219	/* EE  TTY 9 */
#define DECI2_PROTO_EKTTYP	0x021f	/* EE  Kernel TTY */
#define DECI2_PROTO_ESDBGP	0x0230	/* EE  System Debugging */
#define DECI2_PROTO_ETDBGP	0x0240	/* EE  Thread Debugging */
#define DECI2_PROTO_NETMP	0x0400	/* Net Manager */
#define DECI2_PROTO_NTTYP	0x0410	/* Net TTY */

#define DECI2_PROTO_NTTYP	0x0410	/* Net TTY */


#define DECI2_READ			1
#define	DECI2_READDONE		2
#define	DECI2_WRITE			3
#define	DECI2_WRITEDONE		4
#define	DECI2_CHSTATUS		5
#define	DECI2_ERROR			6


typedef struct _tDeci2_DATA {
	_u32 unk0;
	_u32 size_in;
	_u32 unk2;
	_u32 string_left;
	_u32 buff_in_addr;
	_u32 unk5;
	_u32 unk6;
	_u32 unk7;

}Deci2_DATA;


typedef struct _tDeci2_Call {

	_u32	protocol;
	_u32   	data_addr;
	_u32   	handler_addr;


}Deci2_Call;


// deci2 header
typedef struct _tDeci2Hdr{
	_u16 len;	//2
	_u16 rsvd;	//4
	_u16 proto;	//6
	_u8  src;	//7
	_u8  dest;	//8
} Deci2Hdr;


typedef struct _tDeci2_ProtoOPT{
	_s32  	s;			// socket descriptor
	_u32  	wlen;		// write len
	_u32  	rlen;		// read len
	_u32	wptr;		// write buffer pointer
	_u32	rptr;		// read  buffer pointer
	_s32    rbytes;		//???
} Deci2_ProtoOPT;


/* Damn Hack stuff... */
Deci2_Call 		* BiosDeci2;
//Deci2_DATA 		* BiosDeci2Data;

/* Correct ! */
Deci2_ProtoOPT	* BiosDeci2ProtoOPT;
Deci2Hdr		* BiosDeci2Hdr;


_u8 *   BiosDeciBuffer;
_u32	SocketNumber = 0;
_u32	SavedProtoOPTAdd[32];

_u32	Deci2Handler[256];

void SYSCALL_Deci2Call( ) 						{ 
	_u32	Fontion;
//	_u32	Protocol;

	int boucle;

	_u32 * Socket;	

	char temp_string[1024];
	char TTP_string[1024];
	_u32 *   tempTest;

	char * Kputs;
	_u32	Kputs_Add;

/*
#ifdef SYSCALL_LOG
	SYSCALL_LOG("%8.8lx:[BIOS]Deci2Call",(long)MipsCpu.PC );
#endif	
*/
	Fontion = (_u32)MipsCpu.GPR[CPU_REG_A0].lo;
	

	switch( (_u32)Fontion ) {
		//sceDeci2Open
		case 1: 

			BiosDeci2 = ( Deci2_Call * ) &MipsMemory [   (_u32)MipsCpu.GPR[CPU_REG_A1].lo  & 0x01FFFFFF ];
		

			BiosDeci2ProtoOPT 	= ( Deci2_ProtoOPT * ) &MipsMemory [   (_u32)BiosDeci2->data_addr  & 0x01FFFFFF ];
			BiosDeci2Hdr		= ( Deci2Hdr * ) &MipsMemory [   (_u32)BiosDeci2->data_addr  & 0x01FFFFFF ];


#ifdef SYSCALL_LOG
			SYSCALL_LOG("%8.8lx:[BIOS]Deci2CallOpen ( Proto:0x%4.4lx OptAdd:0x%8.8lx EventAdd:0x%8.8lx)",(long)MipsCpu.PC,
				 BiosDeci2->protocol, BiosDeci2->data_addr, BiosDeci2->handler_addr);
#endif	


			SavedProtoOPTAdd[SocketNumber+1] = (_u32)BiosDeci2->data_addr;		

			BiosDeci2ProtoOPT->s = (_s32)++SocketNumber;
			Deci2Handler[ SocketNumber ] = BiosDeci2->handler_addr;	


			MipsCpu.GPR[CPU_REG_V0].lo = BiosDeci2ProtoOPT->s;

		break;


		//sceDeci2Close
		case 2:
#ifdef SYSCALL_LOG
			SYSCALL_LOG("%8.8lx:[BIOS]Deci2CallClose",(long)MipsCpu.PC );
#endif	
		break;

		
		//sceDeci2ReqSend
		case 3:

			BiosDeci2Hdr		= ( Deci2Hdr * ) &MipsMemory [   (_u32)BiosDeci2ProtoOPT->rptr  & 0x01FFFFFF ];

#ifdef SYSCALL_LOG
			SYSCALL_LOG("%8.8lx:[BIOS]Deci2CallReqSend ( sock:0x%8.8lx wlen:0x%8.8lx rlen:0x%8.8lx wptr:0x%8.8lx rptr:0x%8.8lx rbytes:0x%8.8lx )",(long)MipsCpu.PC, 
				BiosDeci2ProtoOPT->s, 	  BiosDeci2ProtoOPT->wlen, 	 
				BiosDeci2ProtoOPT->rlen , BiosDeci2ProtoOPT->wptr ,
				BiosDeci2ProtoOPT->rptr , BiosDeci2ProtoOPT->rbytes 
			);

			SYSCALL_LOG("%8.8lx:[BIOS]Deci2Hdr : len:0x%4.4lx rsvd:0x%4.4lx proto:0x%4.4lx src:%x dst:%x ",(long)MipsCpu.PC, 
				BiosDeci2Hdr->len ,BiosDeci2Hdr->rsvd ,	BiosDeci2Hdr->proto ,BiosDeci2Hdr->src ,
				BiosDeci2Hdr->dest 
			);
#endif	


			if ( BiosDeci2ProtoOPT->s != 0 )  {

			switch ( BiosDeci2Hdr->proto ) {

				case DECI2_PROTO_E0TTYP:
					BiosDeciBuffer = ( _u8 * ) &MipsMemory [ (_u32)(BiosDeci2ProtoOPT->rptr + sizeof(Deci2Hdr) + 4 )  & 0x01FFFFFF ];
					
	
					memcpy( (char*)TTP_string, (char*)BiosDeciBuffer, BiosDeci2Hdr->len - 12 );
					TTP_string[BiosDeci2Hdr->len - 12] = 0;

					sprintf(temp_string,"[TTPY] -> %s",TTP_string);
#ifdef EMU_LOG
					EMU_LOG(temp_string);
#endif

					
//					BIOS_ExecDeci2Handler( Deci2Handler[ BiosDeci2ProtoOPT->s  ] ,DECI2_WRITE, 0, SavedProtoOPTAdd[BiosDeci2ProtoOPT->s] );




 INTC_ExecSubFunc( 	Deci2Handler[ BiosDeci2ProtoOPT->s  ] ,
									INTC_SUBF_SAVE_REG | INTC_SUBF_SET_A0 | INTC_SUBF_SET_A1| INTC_SUBF_SET_A2,
									DECI2_WRITE , 0, SavedProtoOPTAdd[BiosDeci2ProtoOPT->s]);				




				break;


				case DECI2_PROTO_ANY:
					for ( boucle = 1 ; boucle <= SocketNumber ; boucle++ )
//						BIOS_ExecDeci2Handler( Deci2Handler[ boucle  ] ,DECI2_WRITE, 0, SavedProtoOPTAdd[boucle] );



 INTC_ExecSubFunc( 	Deci2Handler[ boucle  ] ,
									INTC_SUBF_SAVE_REG | INTC_SUBF_SET_A0 | INTC_SUBF_SET_A1|INTC_SUBF_SET_A2,
									DECI2_WRITE , 0, SavedProtoOPTAdd[boucle]);				




				break;


				default:
#ifdef EMU_LOG
					EMU_LOG("[Deci2Call] Unknow Protocol : 0x%4.4x",BiosDeci2Hdr->proto);
#endif
				break;
			

			}


			/*Return Value : OK */
			MipsCpu.GPR[CPU_REG_V0].lo = 1;

		} else {


			MipsCpu.GPR[CPU_REG_V0].lo = -1;
		}

		break;

		//sceDeci2Poll
		case 4:

			Socket = ( _u32 * ) &MipsMemory [   (_u32)MipsCpu.GPR[CPU_REG_A1].lo  & 0x01FFFFFF ];
	
#ifdef SYSCALL_LOG
			SYSCALL_LOG("%8.8lx:[BIOS]Deci2CallPoll  sock:0x%8.8lx Handler:0x%8.8lx ",(long)MipsCpu.PC, 
				*Socket /*BiosDeci2ProtoOPT->s*/ , Deci2Handler[ *Socket ]	
			);

#endif	
/*
#ifdef SYSCALL_LOG
			SYSCALL_LOG("%8.8lx:[BIOS]A0:0x%8.8lx A1:0x%8.8lx A2:0x%8.8lx",(long)MipsCpu.PC,
				MipsCpu.GPR[CPU_REG_A0].lo,MipsCpu.GPR[CPU_REG_A1].lo,MipsCpu.GPR[CPU_REG_A2].lo
			 );
#endif	
*/
/*
			BIOS_ExecDeci2Handler( Deci2Handler[ *Socket  ] ,DECI2_WRITEDONE, 0, SavedProtoOPTAdd[ *Socket ] );
			BIOS_ExecDeci2Handler( Deci2Handler[ *Socket  ] ,DECI2_READDONE, 0, SavedProtoOPTAdd[ *Socket ] );
*/




 INTC_ExecSubFunc( 	Deci2Handler[ *Socket  ] ,
									INTC_SUBF_SAVE_REG | INTC_SUBF_SET_A0 | INTC_SUBF_SET_A1| INTC_SUBF_SET_A2,
									DECI2_WRITEDONE , 0, SavedProtoOPTAdd[ *Socket ]);				

/*
 INTC_ExecSubFunc( 	Deci2Handler[ *Socket  ] ,
									INTC_SUBF_SAVE_REG | INTC_SUBF_SET_A0 | INTC_SUBF_SET_A1| INTC_SUBF_SET_A2,
									DECI2_READDONE , 0, SavedProtoOPTAdd[ *Socket ]);				

*/





			//MipsCpu.GPR[CPU_REG_V0].lo = 1;
		break;


		//sceDeci2ExRecv
		case 0xFFFFFFFB:
#ifdef SYSCALL_LOG
			SYSCALL_LOG("%8.8lx:[BIOS]Deci2CallExRecv",(long)MipsCpu.PC );
#endif	
		break;


		//sceDeci2ExSend
		case 0xFFFFFFFA:
#ifdef SYSCALL_LOG
			SYSCALL_LOG("%8.8lx:[BIOS]Deci2CallExSend ( wlen:0x%8.8lx)",(long)MipsCpu.PC,
				BiosDeci2ProtoOPT->wlen
			 );
#endif	
/*
#ifdef SYSCALL_LOG
			SYSCALL_LOG("%8.8lx:[BIOS]A0:0x%8.8lx A1:0x%8.8lx A2:0x%8.8lx",(long)MipsCpu.PC,
				MipsCpu.GPR[CPU_REG_A0].lo,MipsCpu.GPR[CPU_REG_A1].lo,MipsCpu.GPR[CPU_REG_A2].lo
			 );
#endif	
*/
			
			MipsCpu.GPR[CPU_REG_V0].lo = BiosDeci2ProtoOPT->wlen;
		break;


		//sceDeci2ExReqSend
		case 0xFFFFFFF9:
#ifdef SYSCALL_LOG
			SYSCALL_LOG("%8.8lx:[BIOS]Deci2CallExReqSend",(long)MipsCpu.PC );
#endif	
		break;



		//sceDeci2ExLock
		case 0xFFFFFFF8:
#ifdef SYSCALL_LOG
			SYSCALL_LOG("%8.8lx:[BIOS]Deci2CallExLock",(long)MipsCpu.PC );
#endif	
		break;

		
		//sceDeci2ExUnLock
		case 0xFFFFFFF7:
#ifdef SYSCALL_LOG
			SYSCALL_LOG("%8.8lx:[BIOS]Deci2CallExUnLock",(long)MipsCpu.PC );
#endif	
		break;		

		//kputs
		case 0x10:

			Kputs_Add 	= *( (_u32 *) (( _u8 *)( MipsMemory + ( MipsCpu.GPR[CPU_REG_A1].lo & 0x01FFFFFF)) ));
			Kputs 		= ( char * ) &MipsMemory [   (_u32)Kputs_Add  & 0x01FFFFFF ];

#ifdef SYSCALL_LOG
			SYSCALL_LOG("%8.8lx:[BIOS]Deci2Callkputs  Add:0x%8.8lx ",(long)MipsCpu.PC ,
				 (_u32)MipsCpu.GPR[CPU_REG_A1].lo, (_u32)Kputs_Add );
#endif	
					sprintf(temp_string,"[KPUTS] -> %s",Kputs);
#ifdef EMU_LOG
					EMU_LOG(temp_string);
#endif
			
			

		break;

		default:
#ifdef SYSCALL_LOG
			SYSCALL_LOG("%8.8lx:[BIOS]Deci2Call---Unknown",(long)MipsCpu.PC );
#endif	
		break;

	}



    MipsCpu.PC = MipsCpu.PC + 4; 

}     















