/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: atreides.muaddib@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/
#include <stdio.h>
#include <stdlib.h>




#include "PS2Edefs.h"
#include "nSX2.h"

#include "r5900_debug_regs.h"
#include "r5900_main.h"
#include "r5900_core.h"
#include "r5900_core_table.h"
#include "r5900_cop0.h"
#include "r5900_cop0_table.h"
#include "r5900_cop1.h"
#include "r5900_cop1_table.h"
#include "r5900_mmi.h"
#include "r5900_mmi_table.h"

#include "VU_cop2.h"
#include "VU_cop2_table.h"

#include "_Debuger/debuger.h"
 

/* LOG Function for the Interrupt Controller*/
void R5900_OP_Log ( char * ftext  , ...   ) {
	char R5900_OP_LogText[ FILE_NAME_LENGHT ];

	if ( debug_op == 1 ) {
    	va_list ap;
		va_start (ap, ftext);
  		vsprintf (R5900_OP_LogText, ftext, ap);
   		va_end (ap);

		sprintf(debugTempText,"%s\n",R5900_OP_LogText);
	}
}


 
 
/*
Print the the current opcode
*/
void PrintDebugOpcode() {
	
	char 		tempText[100];		//for displaying text to the widget

	if ( MipsCpu.opcode == 0 ) return;



#ifdef EMU_LOG
	EMU_LOG("%8.8lx:(opcode:%8.8lx):", (long)MipsCpu.PC,MipsCpu.opcode);
#endif

	
	switch ( ((MipsCpu.opcode >> 26) & 0x3F) ) {
	
		case 0 	: 	sprintf(tempText,"r5900:dbg_opcode_special: %s \n", debug_special_opc[(MipsCpu.opcode  & 0x3F)]); break;
		case 1	: 	sprintf(tempText,"r5900:dbg_opcode_regimm: %s \n", debug_regimm_opc[(MipsCpu.opcode  & 0x1F0000)>>16]); break;
		case 16 : 	sprintf(tempText,"r5900:dbg_opcode_cop0: %s \n", debug_cop0_opc[(MipsCpu.opcode&0x3E00000)>>21]);


#ifdef EMU_LOG
	EMU_LOG(tempText);
#endif

				  	sprintf(tempText,"r5900:dbg_opcode_cop0c0: %s \n", debug_cop0c0_opc[(MipsCpu.opcode&0x3F)]);
					break;
		case 17 : {
					switch( (MipsCpu.opcode&0x3E00000)>>21) {
						
						case 8:  sprintf(tempText,"r5900:debug_cop1_BC1_opc: %s \n",	debug_cop1_bc1_opc[(MipsCpu.opcode>>16)&0x3] );break;					
						case 16: sprintf(tempText,"r5900:debug_cop1_S_opc: %s \n", 	debug_cop1_s_opc[MipsCpu.opcode&0x3F] 	);break;					
						case 20: sprintf(tempText,"r5900:debug_cop1_W_opc: %s \n", 	debug_cop1_W_L_opc[MipsCpu.opcode&0x3F] );break;					

						default: sprintf(tempText,"r5900:dbg_opcode_cop1: %s \n", debug_cop1_opc[(MipsCpu.opcode&0x3E00000)>>21]);break;					
					}
					
					break;
				}	
		case 18	: {
					if ( ((MipsCpu.opcode&0x3E00000)>>21) > 15 ) {
						

						switch ( MipsCpu.opcode&0x3F ) {

							case 60 : sprintf(tempText,"r5900:debug_cop2_V_t60_opc: %s \n", debug_cop2_V_t60_opc[(MipsCpu.opcode & 0x7C0)>>6]);break;
							case 61 : sprintf(tempText,"r5900:debug_cop2_V_t61_opc: %s \n", debug_cop2_V_t61_opc[(MipsCpu.opcode & 0x7C0)>>6]);break;
							case 62 : sprintf(tempText,"r5900:debug_cop2_V_t62_opc: %s \n", debug_cop2_V_t62_opc[(MipsCpu.opcode & 0x7C0)>>6]);break;
							case 63: sprintf(tempText,"r5900:debug_cop2_V_t63_opc: %s \n", debug_cop2_V_t63_opc[(MipsCpu.opcode & 0x7C0)>>6]);break;
							default : sprintf(tempText,"r5900:debug_pcop2_Vopc: %s \n", debug_pcop2_V_opc[MipsCpu.opcode&0x3F]);break;
						}


					} else { 

						sprintf(tempText,"r5900:debug_pcop2_opc: %s \n", debug_pcop2_opc[(MipsCpu.opcode&0x3E00000)>>21]);
					}

				}
				break;

		case 28 : {

				switch  (MipsCpu.opcode  & 0x3F) {
			
				case 8:sprintf(tempText,"r5900:dbg_opcode_special2MMI0: %s \n", debug_Special2_MMI0[((_OPCODE_ >> 6)  & 0x1F)] );break;
				case 9:sprintf(tempText,"r5900:dbg_opcode_special2MMI2: %s \n", debug_Special2_MMI2[((_OPCODE_ >> 6)  & 0x1F)] );break;

				case 40:sprintf(tempText,"r5900:dbg_opcode_special2MMI1: %s \n", debug_Special2_MMI1[((_OPCODE_ >> 6)  & 0x1F)] );break;
				case 41:sprintf(tempText,"r5900:dbg_opcode_special2MMI3: %s \n", debug_Special2_MMI3[((_OPCODE_ >> 6)  & 0x1F)] );break;
				
				default : sprintf(tempText,"r5900:dbg_opcode_special2: %s \n", debug_special2_opc[((_OPCODE_)  & 0x3F)] );
				}			

			}
			break;


		
	
		default : sprintf(tempText,"r5900:dbg_opcode: %s \n", debug_normal_opc[((MipsCpu.opcode >> 26) & 0x3F)]); break;
	
	}
#ifdef EMU_LOG
	EMU_LOG(tempText);
#endif
}


/* Main CPU function*/
void MipsUpdateCpu ( ) {

	/*Update the Number of Instructions*/
	MipsCpu.Nb_instructions++;		

/*
	if ( (_u64)MipsCpu.Nb_instructions == (_u64)454650 ) {
			MipsCpu.halt = 1;
			(* DBG_SendBreak)( );
	}
*/
/*
	if ( (_u64)MipsCpu.PC == (_u64)0x001AD768 ) {
			MipsCpu.halt = 1;
			(* DBG_SendBreak)( );
	}
*/

	/* COP0 update : need some check and fix */
	MipsCpu.CPR0[9][0] =	MipsCpu.CPR0[9][0] + 1;

	
	/* read the Memory at PC and Set MipsCpu.opcode */
	
	/* Check if PC is inside BIOS area or Main Area 
		 !!! Need to find a better way !!!
	*/
	if ( (( (_u32)MipsCpu.PC >= (_u32)0x1FC00000 ) && ( (_u32)MipsCpu.PC <= (_u32)0x1FFFFFFF )) ||
		 (( (_u32)MipsCpu.PC >= (_u32)0x9FC00000 ) && ( (_u32)MipsCpu.PC <= (_u32)0x9FFFFFFF )) ||
		 (( (_u32)MipsCpu.PC >= (_u32)0xBFC00000 ) && ( (_u32)MipsCpu.PC <= (_u32)0xBFFFFFFF )) 
		)
		MipsCpu.opcode=*( (_u64 *) (( _u8 *)( BiosMemory + ( MipsCpu.PC & 0x3FFFFF)) ));
	else
	 	MipsCpu.opcode=*( (_u64 *) (( _u8 *)( MipsMemory + ( MipsCpu.PC & 0x01FFFFFF)) ));



	//MipsCpu.Nb_Cycles =  (_u64)((_u64)MipsCpu.Nb_Cycles + Timers_GetCycles(_OPCODE_));
	MipsCpu.Nb_Cycles  = MipsCpu.Nb_Cycles  + 4;

	if ( _OPCODE_ != 0 ) normal_opc[((_OPCODE_ >> 26) & 0x3F)]();	//Execute the current opcode
	else NOP();

}











