/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: atreides.muaddib@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/
#include <stdio.h>
#include <stdlib.h>

#include <strings.h>
#include <float.h>
#include <math.h>

#include "nSX2.h"
 
#include "r5900_main.h"

#include "r5900_debug_regs.h"

#include "r5900_core.h"	//for NI

#include "r5900_mmi.h"
#include "r5900_mmi_table.h"


#include "PS2Edefs.h"


/*
POFIS opcodes :
MADD1, MADDU, MADDU1, MFSA, MTSA, MTSAB, MULTU1
*/

 


/*****************************************************************************************************************
								EE SPECIFIC INSTRUCTION SET
*****************************************************************************************************************/

void    SPECIAL2()	{	special2_opc[(_OPCODE_  & 0x3F)](); }

void 	MMI0()		{	Special2_MMI0[((_OPCODE_ >> 6)  & 0x1F)]();}
 
void 	MMI2()		{	Special2_MMI2[((_OPCODE_ >> 6)  & 0x1F)]();}

void 	MMI1()		{	Special2_MMI1[((_OPCODE_ >> 6)  & 0x1F)]();}

void 	MMI3()		{	Special2_MMI3[((_OPCODE_ >> 6)  & 0x1F)]();}




void	DIV1	( void )	{		

	if ( (_s32)MipsCpu.GPR[CURRENT_RT].lo != 0 ) {
		MipsCpu.GPR[CPU_REG_LO].hi = (_s64)((_s32)MipsCpu.GPR[CURRENT_RS].lo / (_s32)MipsCpu.GPR[CURRENT_RT].lo);
    	MipsCpu.GPR[CPU_REG_HI].hi = (_s64)((_s32)MipsCpu.GPR[CURRENT_RS].lo % (_s32)MipsCpu.GPR[CURRENT_RT].lo);
	}
	
	INCREMENT_PC();
}
 

void	DIVU1	( void )	{	

	if ( (_s32)MipsCpu.GPR[CURRENT_RT].lo != 0 ) {
		MipsCpu.GPR[CPU_REG_LO].hi = (_u64)((_u32)MipsCpu.GPR[CURRENT_RS].lo / (_u32)MipsCpu.GPR[CURRENT_RT].lo);
    	MipsCpu.GPR[CPU_REG_HI].hi = (_u64)((_u32)MipsCpu.GPR[CURRENT_RS].lo % (_u32)MipsCpu.GPR[CURRENT_RT].lo);	
	}
	
	INCREMENT_PC();
} 


void    LQ		( void )	{	
	_u64 off;

	off = Extractoffset();
	sign_extendedh(off);


	MipsMemoryRead ( &MipsCpu.GPR[CURRENT_RT] , DATA_QUAD ,  MipsCpu.GPR[CURRENT_BASE].lo , off ) ;

    INCREMENT_PC();
}


void	MADD	( void )	{	
	_s64 temp;


    temp = (_s64)((u64)MipsCpu.GPR[CPU_REG_LO].lo | ((u64)MipsCpu.GPR[CPU_REG_HI].lo << 32)) + 
               (_s64)((_s32)MipsCpu.GPR[CURRENT_RS].lo * (_s32)MipsCpu.GPR[CURRENT_RT].lo);
	
	MipsCpu.GPR[CPU_REG_LO].lo = (_s32)( temp & 0xffffffff);
	MipsCpu.GPR[CPU_REG_HI].lo = (_s32)( (_s64)temp  >> 32 );

	if ( CURRENT_RD != 0 ) MipsCpu.GPR[CURRENT_RD].lo = MipsCpu.GPR[CPU_REG_LO].lo;

	INCREMENT_PC();
} 


void	MADD1	( void )	{		
	_s64 temp;


    temp = (_s64)((u64)MipsCpu.GPR[CPU_REG_LO].hi | ((u64)MipsCpu.GPR[CPU_REG_HI].hi << 32)) + 
           (_s64)((_s32)MipsCpu.GPR[CURRENT_RS].lo * (_s32)MipsCpu.GPR[CURRENT_RT].lo);
	
	MipsCpu.GPR[CPU_REG_LO].hi = (_s32)( temp & 0xffffffff);
	MipsCpu.GPR[CPU_REG_HI].hi = (_s32)( (_s64)temp  >> 32 );


	if ( CURRENT_RD != 0 ) MipsCpu.GPR[CURRENT_RD].lo = MipsCpu.GPR[CPU_REG_LO].lo;

	INCREMENT_PC();
} 


void	MADDU	( void )	{		
	_u64 temp;


   temp = (_u64)((u64)MipsCpu.GPR[CPU_REG_LO].lo | ((u64)MipsCpu.GPR[CPU_REG_HI].lo << 32)) + 
               (_u64)((_s32)MipsCpu.GPR[CURRENT_RS].lo * (_s32)MipsCpu.GPR[CURRENT_RT].lo);
	
	MipsCpu.GPR[CPU_REG_LO].lo = (_s32)( temp & 0xffffffff);
	MipsCpu.GPR[CPU_REG_HI].lo = (_s32)( (_s64)temp  >> 32 );


	if ( CURRENT_RD != 0 ) MipsCpu.GPR[CURRENT_RD].lo = MipsCpu.GPR[CPU_REG_LO].lo;

	INCREMENT_PC();

} 

void	MADDU1	( void )	{		
	_u64 temp;

   temp = (_u64)((u64)MipsCpu.GPR[CPU_REG_LO].hi | ((u64)MipsCpu.GPR[CPU_REG_HI].hi << 32)) + 
               (_u64)((_s32)MipsCpu.GPR[CURRENT_RS].lo * (_s32)MipsCpu.GPR[CURRENT_RT].lo);
	
	MipsCpu.GPR[CPU_REG_LO].hi = (_s32)( temp & 0xffffffff);
	MipsCpu.GPR[CPU_REG_HI].hi = (_s32)( (_s64)temp  >> 32 );

	if ( CURRENT_RD != 0 ) MipsCpu.GPR[CURRENT_RD].lo = MipsCpu.GPR[CPU_REG_LO].lo;

	INCREMENT_PC();
}  


void 	MFHI1	( void )	{	
	
	MipsCpu.GPR[CURRENT_RD].lo = (_u64)MipsCpu.GPR[CPU_REG_HI].hi;

    INCREMENT_PC();
}


void 	MFLO1	( void )	{	

	
	MipsCpu.GPR[CURRENT_RD].lo = (_u64)MipsCpu.GPR[CPU_REG_LO].hi;

    INCREMENT_PC();


}


void 	MFSA	( void )	{		


	MipsCpu.GPR[CURRENT_RD].lo = (_u64)MipsCpu.SA.lo;
	INCREMENT_PC();
}


void 	MTHI1	( void )	{		

	
	MipsCpu.GPR[CPU_REG_HI].hi = (_u64)MipsCpu.GPR[CURRENT_RS].lo ;

    INCREMENT_PC();
}

void 	MTLO1	( void )	{	

	
	MipsCpu.GPR[CPU_REG_LO].hi = (_u64)MipsCpu.GPR[CURRENT_RS].lo ;

    INCREMENT_PC();
} 


void 	MTSA	( void )	{		


	MipsCpu.SA.lo = (_u64)MipsCpu.GPR[CURRENT_RS].lo;
	INCREMENT_PC();

}


void 	MTSAB	( void )	{		


	MipsCpu.SA.lo =  (((_u32)(MipsCpu.GPR[CURRENT_RS].lo ) & 0xF ) ^ ( CURRENT_IMM & 0xF ) ) * 8;
	INCREMENT_PC();

}


void 	MTSAH 	( void )	{		


	MipsCpu.SA.lo =  (((_s32)(MipsCpu.GPR[CURRENT_RS].lo ) & 0x7 ) ^ ( CURRENT_IMM & 0x7 ) );
	
	INCREMENT_PC();
}


void	MULT1	( void )	{	
	_s64 temp;


	
	temp = (_s64)((_s32)(MipsCpu.GPR[CURRENT_RS].lo )) * ((_s32)( MipsCpu.GPR[CURRENT_RT].lo ));
	
	MipsCpu.GPR[CPU_REG_LO].hi = (_s64)(temp & 0xffffffff);
	MipsCpu.GPR[CPU_REG_HI].hi = (_s64)((_s64)temp  >> 32 );

	MipsCpu.GPR[CURRENT_RD].lo = (_u64)MipsCpu.GPR[CPU_REG_LO].hi ;
	
	INCREMENT_PC();
}



void 	MULTU1	( void )	{	
    _u64 temp;

	temp = (_u64)((_u32)(MipsCpu.GPR[CURRENT_RS].lo )) * ((_u32)( MipsCpu.GPR[CURRENT_RT].lo ));
	
	MipsCpu.GPR[CPU_REG_LO].hi = (_s64)((_s32)(temp & 0xffffffff));
	MipsCpu.GPR[CPU_REG_HI].hi = (_s64)(_s32)(((_s64)temp  >> 32 ));

	if ( CURRENT_RD != 0 ) MipsCpu.GPR[CURRENT_RD].lo = (_u64)MipsCpu.GPR[CPU_REG_LO].hi ;
	
	INCREMENT_PC();

} 


void 	PABSH	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;}
void 	PABSW	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 
void 	PADDB	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 
void 	PADDH	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 
void 	PADDSB	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 
void 	PADDSW	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 
void 	PADDSH	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 


#define dPADDUB( fval )  \
temp16 = (_u16)((_u8)_valRS->b8.fval + (_u8)_valRT->b8.fval ); \
if ( (_u16)temp16 > (_u16)0xFF ) _valRD->b8.fval = (_u8)0xFF; \
else _valRD->b8.fval = (_u8)temp16;  
void 	PADDUB	( void )	{		

	_u16	temp16;

	u128v2 * _valRT;
	u128v2 * _valRD;
	u128v2 * _valRS;

	
	_valRD = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RD] ;	
	_valRS = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RS] ;
	_valRT = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RT] ;

	dPADDUB(b00);dPADDUB(b01);dPADDUB(b02);dPADDUB(b03);
	dPADDUB(b04);dPADDUB(b05);dPADDUB(b06);dPADDUB(b07);
	dPADDUB(b08);dPADDUB(b09);dPADDUB(b10);dPADDUB(b11);
	dPADDUB(b12);dPADDUB(b13);dPADDUB(b14);dPADDUB(b15);

	INCREMENT_PC();
} 


void 	PADDUH	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;

} 


#define dPADDUW( fval ) \
temp64 = (_u64)((_u32)_valRS->b32.fval + (_u32)_valRT->b32.fval ); \
if ( (_u64)temp64 > (_u64)0xFFFFFFFF ) _valRD->b32.fval = (_u32)0xFFFFFFFF; \
else _valRD->b32.fval = (_u32)temp64; 
void 	PADDUW	( void )	{	

	_u64	temp64;

	u128v2 * _valRT;
	u128v2 * _valRD;
	u128v2 * _valRS;

	
	_valRD = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RD] ;	
	_valRS = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RS] ;
	_valRT = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RT] ;

	dPADDUW(w00);dPADDUW(w01);dPADDUW(w02);dPADDUW(w03);

	INCREMENT_PC();
} 


#define dPADDW(fval) _valRD->b32.fval = (_s32)((_s32)_valRS->b32.fval + (_s32)_valRT->b32.fval);
void 	PADDW	( void )	{		

	u128v2 * _valRT;
	u128v2 * _valRD;
	u128v2 * _valRS;

	_valRD = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RD] ;	
	_valRS = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RS] ;
	_valRT = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RT] ;

	dPADDW(w00);dPADDW(w01);dPADDW(w02);dPADDW(w03);

	INCREMENT_PC();
}  


void 	PADSBH	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 

void 	PAND	( void )	{	

	MipsCpu.GPR[CURRENT_RD].lo = (_u64)((_u64)MipsCpu.GPR[CURRENT_RS].lo & (_u64)MipsCpu.GPR[CURRENT_RT].lo);
	MipsCpu.GPR[CURRENT_RD].hi = (_u64)((_u64)MipsCpu.GPR[CURRENT_RS].hi & (_u64)MipsCpu.GPR[CURRENT_RT].hi);
	
	INCREMENT_PC();
}  


void 	PCEQH	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 
void 	PCEQW	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 


#define dPCEQB( fval ) \
if ( (_u8)_valRS->b8.fval == (_u8)_valRT->b8.fval ) _valRD->b8.fval = (_u8)0xFF; \
else (_u8)_valRD->b8.fval = (_u8)0; 
void 	PCEQB	( void )	{		

	u128v2 * _valRT;
	u128v2 * _valRD;
	u128v2 * _valRS;
	
	_valRD = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RD] ;	
	_valRS = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RS] ;
	_valRT = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RT] ;

	dPCEQB(b00);dPCEQB(b01);dPCEQB(b02);dPCEQB(b03);
	dPCEQB(b04);dPCEQB(b05);dPCEQB(b06);dPCEQB(b07);
	dPCEQB(b08);dPCEQB(b09);dPCEQB(b10);dPCEQB(b11);
	dPCEQB(b12);dPCEQB(b13);dPCEQB(b14);dPCEQB(b15);

	INCREMENT_PC();
} 


void 	PCGTB	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 
void 	PCGTH	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 
void 	PCGTW	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 

 
void 	PCPYH	( void )	{	
	u128v2 * _valRT;
	u128v2 * _valRD;

	
	_valRT = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RT] ;
	_valRD = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RD] ;


	_valRD->b16.h00 = (_u16)_valRT->b16.h00;
	_valRD->b16.h01 = (_u16)_valRT->b16.h00;
	_valRD->b16.h02 = (_u16)_valRT->b16.h00;
	_valRD->b16.h03 = (_u16)_valRT->b16.h00;
	_valRD->b16.h04 = (_u16)_valRT->b16.h04;
	_valRD->b16.h05 = (_u16)_valRT->b16.h04;
	_valRD->b16.h06 = (_u16)_valRT->b16.h04;
	_valRD->b16.h07 = (_u16)_valRT->b16.h04;

	INCREMENT_PC();
} 


void 	PCPYLD	( void )	{	

	MipsCpu.GPR[CURRENT_RD].lo = (_u64)MipsCpu.GPR[CURRENT_RT].lo;
	MipsCpu.GPR[CURRENT_RD].hi = (_u64)MipsCpu.GPR[CURRENT_RS].lo;

	INCREMENT_PC();

} 

void 	PCPYUD	( void )	{	


	MipsCpu.GPR[CURRENT_RD].lo = (_u64)MipsCpu.GPR[CURRENT_RS].hi;
	MipsCpu.GPR[CURRENT_RD].hi = (_u64)MipsCpu.GPR[CURRENT_RT].hi;

	INCREMENT_PC();

} 

void 	PDIVBW	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 
void 	PDIVUW	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 
void 	PDIVW	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 
void 	PEXCH	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 
void 	PEXCW	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;}  
void 	PEXEH	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;}   

void 	PEXEW	( void )	{	

	u128v2 * _valRT;
	u128v2 * _valRD;

	_u32	savedX;

	_valRT = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RT] ;
	_valRD = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RD] ;


	_valRD->b32.w00 = (_u32)_valRT->b32.w02;
	_valRD->b32.w01 = (_u32)_valRT->b32.w01;
	_valRD->b32.w02 = (_u32)_valRT->b32.w00;
	_valRD->b32.w03 = (_u32)_valRT->b32.w03;

	INCREMENT_PC();
} 


void 	PEXT5	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 

void 	PEXTLB	( void )	{	

	u128v2 * _valRT;
	u128v2 * _valRD;
	u128v2 * _valRS;
	
	
	_valRT = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RT] ;
	_valRD = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RD] ;
	_valRS = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RS] ;


	_valRD->b8.b00 = (_u8)_valRT->b8.b00;
	_valRD->b8.b01 = (_u8)_valRS->b8.b00;
	_valRD->b8.b02 = (_u8)_valRT->b8.b01;
	_valRD->b8.b03 = (_u8)_valRS->b8.b01;
	_valRD->b8.b04 = (_u8)_valRT->b8.b02;
	_valRD->b8.b05 = (_u8)_valRS->b8.b02;
	_valRD->b8.b06 = (_u8)_valRT->b8.b03;
	_valRD->b8.b07 = (_u8)_valRS->b8.b03;
	_valRD->b8.b08 = (_u8)_valRT->b8.b04;
	_valRD->b8.b09 = (_u8)_valRS->b8.b04;
	_valRD->b8.b10 = (_u8)_valRT->b8.b05;
	_valRD->b8.b11 = (_u8)_valRS->b8.b05;
	_valRD->b8.b12 = (_u8)_valRT->b8.b06;
	_valRD->b8.b13 = (_u8)_valRS->b8.b06;
	_valRD->b8.b14 = (_u8)_valRT->b8.b07;
	_valRD->b8.b15 = (_u8)_valRS->b8.b07;

	INCREMENT_PC();


}  

void 	PEXTLH		( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;}  

void 	PEXTLW	( void )	{		
	u128v2 * _valRT;
	u128v2 * _valRD;
	u128v2 * _valRS;
	
	_valRT = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RT] ;
	_valRD = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RD] ;
	_valRS = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RS] ;

	_valRD->b32.w00 = (_u32)_valRT->b32.w00;
	_valRD->b32.w01 = (_u32)_valRS->b32.w00;
	_valRD->b32.w02 = (_u32)_valRT->b32.w01;
	_valRD->b32.w03 = (_u32)_valRS->b32.w01;

	INCREMENT_PC();
}  


void 	PEXTUB	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 
void 	PEXTUH	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 



void 	PEXTUW	( void )	{	


	u128v2 * _valRT;
	u128v2 * _valRD;
	u128v2 * _valRS;
	
	_valRT = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RT] ;
	_valRD = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RD] ;
	_valRS = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RS] ;

	_valRD->b32.w00 = _valRT->b32.w02;
	_valRD->b32.w01 = _valRS->b32.w02;
	_valRD->b32.w02 = _valRT->b32.w03;
	_valRD->b32.w03 = _valRS->b32.w03;

	INCREMENT_PC();
} 


void 	PHMADH	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 
void 	PHMSBH	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;}
void 	PINTH	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 
void 	PINTEH	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 



#define _PLZCW(fval) { \
    int i; \
	u32 sign; \
 \
 \
	sign = _valRS->b32.fval >> 31; \
    for (i=30; i>=0; i--) { \
        if (((_valRS->b32.fval >> i) & 0x1) != sign) { \
            break; \
        } \
    } \
    _valRD->b32.fval = 30 - i; \
}
void	PLZCW	( void )	{	
	_u64	temp64;

	u128v2 * _valRT;
	u128v2 * _valRD;
	u128v2 * _valRS;

	_valRD = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RD] ;	
	_valRS = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RS] ;
	_valRT = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RT] ;


	_PLZCW(w00)	;	_PLZCW(w01)	;
	INCREMENT_PC();
} 


void 	PMADDH	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 
void 	PMADDUW	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 
void 	PMADDW	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 
void 	PMAXW	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 
void 	PMAXH	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;}  
void 	PMFHI	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 


void 	PMFHL	( void )	{	
	_u32	fmt;

	u128v2 * _valRD;
	u128v2 * _valHI;
	u128v2 * _valLO;

	
	_valRD = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RD] ;	
	_valHI = ( u128v2 * ) &MipsCpu.GPR[CPU_REG_HI] ;
	_valLO = ( u128v2 * ) &MipsCpu.GPR[CPU_REG_LO] ;



	fmt = CURRENT_SA;

	switch ( fmt ) {
	
	case 3:

		_valRD->b16.h00 = _valLO->b16.h00;
		_valRD->b16.h01 = _valLO->b16.h02;
		_valRD->b16.h02 = _valHI->b16.h00;
		_valRD->b16.h03 = _valHI->b16.h02;
		_valRD->b16.h04 = _valLO->b16.h04;
		_valRD->b16.h05 = _valLO->b16.h06;
		_valRD->b16.h06 = _valHI->b16.h04;
		_valRD->b16.h07 = _valHI->b16.h06;

	break;


	default :
		EMU_LOG("PMFHL , case NI %d",fmt);
	break;	
	}

	INCREMENT_PC();

	/*MipsCpu.halt = 1;*/
} 




void 	PMFLO	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 
void 	PMINW	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;}  
void 	PMINH	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 
void 	PMSUBH	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;}  
void 	PMSUBW	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 


void 	PMTHI	( void )	{		

	MipsCpu.GPR[CPU_REG_HI].lo = (_u64)MipsCpu.GPR[CURRENT_RS].lo;
	MipsCpu.GPR[CPU_REG_HI].hi = (_u64)MipsCpu.GPR[CURRENT_RS].hi;

	INCREMENT_PC();
} 


void 	PMTHL	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 


void 	PMTLO	( void )	{	

	MipsCpu.GPR[CPU_REG_LO].lo = (_u64)MipsCpu.GPR[CURRENT_RS].lo;
	MipsCpu.GPR[CPU_REG_LO].hi = (_u64)MipsCpu.GPR[CURRENT_RS].hi;

	INCREMENT_PC();

} 


void 	PMULTH	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;}  
void 	PMULTUW	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 
void 	PMULTW	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 


void 	PNOR	( void )	{		
	
	MipsCpu.GPR[CURRENT_RD].lo = (_u64)~((_u64)MipsCpu.GPR[CURRENT_RS].lo | (_u64)MipsCpu.GPR[CURRENT_RT].lo);
	MipsCpu.GPR[CURRENT_RD].hi = (_u64)~((_u64)MipsCpu.GPR[CURRENT_RS].hi | (_u64)MipsCpu.GPR[CURRENT_RT].hi);
	
	INCREMENT_PC();
} 


void 	POR		( void )	{	

	MipsCpu.GPR[CURRENT_RD].lo = (_u64)((_u64)MipsCpu.GPR[CURRENT_RS].lo | (_u64)MipsCpu.GPR[CURRENT_RT].lo);
	MipsCpu.GPR[CURRENT_RD].hi = (_u64)((_u64)MipsCpu.GPR[CURRENT_RS].hi | (_u64)MipsCpu.GPR[CURRENT_RT].hi);
	
	INCREMENT_PC();

} 


#define _PPAC5(fval) \
	_valRD->b32.fval = (_u32)\
		((_valRT->b32.fval >>  3) & 0x0000001F) | \
		((_valRT->b32.fval) & 0x000003E0) | \
		((_valRT->b32.fval) & 0x00007C00) | \
		((_valRT->b32.fval) & 0x00008000);

void 	PPAC5	( void )	{	
	_u64	temp64;

	u128v2 * _valRT;
	u128v2 * _valRD;
	u128v2 * _valRS;

	
	_valRD = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RD] ;	
	_valRS = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RS] ;
	_valRT = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RT] ;

	_PPAC5(w00);_PPAC5(w01);_PPAC5(w02);_PPAC5(w03);

	INCREMENT_PC();
}  



void 	PPACB	( void )	{	

	u128v2 * _valRT;
	u128v2 * _valRD;
	u128v2 * _valRS;

	
	
	_valRT = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RT] ;
	_valRD = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RD] ;
	_valRS = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RS] ;


	_valRD->b8.b00 = (_u8)_valRT->b8.b00;
	_valRD->b8.b01 = (_u8)_valRT->b8.b02;
	_valRD->b8.b02 = (_u8)_valRT->b8.b04;
	_valRD->b8.b03 = (_u8)_valRT->b8.b06;
	_valRD->b8.b04 = (_u8)_valRT->b8.b08;
	_valRD->b8.b05 = (_u8)_valRT->b8.b10;
	_valRD->b8.b06 = (_u8)_valRT->b8.b12;
	_valRD->b8.b07 = (_u8)_valRT->b8.b14;
	_valRD->b8.b08 = (_u8)_valRS->b8.b00;
	_valRD->b8.b09 = (_u8)_valRS->b8.b02;
	_valRD->b8.b10 = (_u8)_valRS->b8.b04;
	_valRD->b8.b11 = (_u8)_valRS->b8.b06;
	_valRD->b8.b12 = (_u8)_valRS->b8.b08;
	_valRD->b8.b13 = (_u8)_valRS->b8.b10;
	_valRD->b8.b14 = (_u8)_valRS->b8.b12;
	_valRD->b8.b15 = (_u8)_valRS->b8.b14;


	INCREMENT_PC();


}  


void 	PPACH	( void )	{	
	u128v2 * _valRT;
	u128v2 * _valRD;
	u128v2 * _valRS;

	
	
	_valRT = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RT] ;
	_valRD = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RD] ;
	_valRS = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RS] ;

	_valRD->b16.h00 = (_u16)_valRT->b16.h00;
	_valRD->b16.h01 = (_u16)_valRT->b16.h02;
	_valRD->b16.h02 = (_u16)_valRT->b16.h04;
	_valRD->b16.h03 = (_u16)_valRT->b16.h06;
	_valRD->b16.h04 = (_u16)_valRS->b16.h00;
	_valRD->b16.h05 = (_u16)_valRS->b16.h02;
	_valRD->b16.h06 = (_u16)_valRS->b16.h04;
	_valRD->b16.h07 = (_u16)_valRS->b16.h06;

	INCREMENT_PC();
}   


void 	PPACW	( void )	{		

	u128v2 * _valRT;
	u128v2 * _valRD;
	u128v2 * _valRS;

	
	_valRT = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RT] ;
	_valRD = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RD] ;
	_valRS = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RS] ;

	_valRD->b32.w00 = _valRT->b32.w00;
	_valRD->b32.w01 = _valRT->b32.w02;
	_valRD->b32.w02 = _valRS->b32.w00;
	_valRD->b32.w03 = _valRS->b32.w02;

	INCREMENT_PC();

}  


void 	PREVH	( void )	{		

	u128v2 * _valRT;
	u128v2 * _valRD;
	u128v2 * _valRS;

	
	_valRT = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RT] ;
	_valRD = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RD] ;
	_valRS = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RS] ;


	_valRD->b16.h00 = (_u16)_valRT->b16.h03;
	_valRD->b16.h01 = (_u16)_valRT->b16.h02;
	_valRD->b16.h02 = (_u16)_valRT->b16.h01;
	_valRD->b16.h03 = (_u16)_valRT->b16.h00;

	_valRD->b16.h04 = (_u16)_valRT->b16.h07;
	_valRD->b16.h05 = (_u16)_valRT->b16.h06;
	_valRD->b16.h06 = (_u16)_valRT->b16.h05;
	_valRD->b16.h07 = (_u16)_valRT->b16.h04;



	INCREMENT_PC();
}  


void 	PROT3W	( void )	{	

	u128v2 * _valRT;
	u128v2 * _valRD;

	_u32	savedX;

	
	
	_valRT = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RT] ;
	_valRD = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RD] ;

	savedX = _valRD->b32.w00;

	_valRD->b32.w00 = (_u32)_valRT->b32.w01;
	_valRD->b32.w01 = (_u32)_valRT->b32.w02;
	_valRD->b32.w02 = (_u32)savedX;
	_valRD->b32.w03 = (_u32)_valRT->b32.w03;

	INCREMENT_PC();

} 


void 	PSLLH	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 
void 	PSLLVW	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 
void 	PSLLW	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;}
void 	PSRAH	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 
void 	PSRAVW	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 
void 	PSRAW	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;}
void 	PSRLH	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 
void 	PSRLVW	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 
void 	PSRLW	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 
 

#define	dPSUBB(fval)	_valRD->b8.fval = (_s8)((_s8)_valRS->b8.fval - (_s8)_valRT->b8.fval);
void 	PSUBB	( void )	{		

	u128v2 * _valRT;
	u128v2 * _valRD;
	u128v2 * _valRS;


	_valRD = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RD] ;	
	_valRS = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RS] ;
	_valRT = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RT] ;

	dPSUBB(b00);dPSUBB(b01);dPSUBB(b02);dPSUBB(b03);
	dPSUBB(b04);dPSUBB(b05);dPSUBB(b06);dPSUBB(b07);
	dPSUBB(b08);dPSUBB(b09);dPSUBB(b10);dPSUBB(b11);
	dPSUBB(b12);dPSUBB(b13);dPSUBB(b14);dPSUBB(b15);

	INCREMENT_PC();

}  

void 	PSUBH	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;}  
void 	PSUBSB	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;}  
void 	PSUBSW	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;}  
void 	PSUBSH	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;}  
void 	PSUBUB	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 
void 	PSUBUH	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 
void 	PSUBUW	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 



#define dPSUBW(fval) _valRD->b32.fval = (_s32)((_s32)_valRS->b32.fval - (_s32)_valRT->b32.fval);
void 	PSUBW	( void )	{	
	u128v2 * _valRT;
	u128v2 * _valRD;
	u128v2 * _valRS;


	_valRD = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RD] ;	
	_valRS = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RS] ;
	_valRT = ( u128v2 * ) &MipsCpu.GPR[CURRENT_RT] ;

	dPSUBW(w00);dPSUBW(w01);dPSUBW(w02);dPSUBW(w03);


	INCREMENT_PC();
} 


void 	PXOR	( void )	{		


	MipsCpu.GPR[CURRENT_RD].lo = (_u64)((_u64)MipsCpu.GPR[CURRENT_RS].lo ^ (_u64)MipsCpu.GPR[CURRENT_RT].lo);
	MipsCpu.GPR[CURRENT_RD].hi = (_u64)((_u64)MipsCpu.GPR[CURRENT_RS].hi ^ (_u64)MipsCpu.GPR[CURRENT_RT].hi);
	
	INCREMENT_PC();
}  


void 	QFSRV	( void )	{		PrintDebugOpcode() ; MipsCpu.halt = 1;} 


void 	SQ	( void )	{	
	_u64 off;

	
	off = Extractoffset();
	sign_extendedh(off);
	
	if ( debug_op == 1 ) 	sprintf(debugTempText,"%8.8llx: sq %s,%4llx(%s)", (long)MipsCpu.PC, debug_r5900_reg[CURRENT_RT],CURRENT_OFFSET,debug_r5900_reg[CURRENT_BASE] );

	MipsMemoryWrite ( MipsCpu.GPR[CURRENT_RT] , DATA_QUAD ,  MipsCpu.GPR[CURRENT_BASE].lo , off ) ;
    INCREMENT_PC();
}

