/**
 * nSX2 - an OpenSource PS2 Emulator
 * Copyright (C) 2002 Muad
 *
 * homepage: http://nsx2.emulation64.com
 * email address: atreides.muaddib@caramail.com
 * 
 * If you want to contribute to the project please contact
 * me first (maybe someone is already making what you are
 * planning to do).
 *
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
**/
#include <stdlib.h>
#include <stdio.h>

#include "nSX2.h"
#include "PS2Edefs.h"

#include "timers.h"

STimer Timer[ 8 ];

TimersReg	TimerGSReg;
TimersReg	TimerInputReg;
TimersReg	TimerMainReg;
TimersReg	TimerThreadReg;

_u64	TIMERS_PreviousCycles;



void TIMERS_Log ( char * ftext  , ...   ) {
	char TIMER_LogText[ FILE_NAME_LENGHT ];

	if ( debug_timer == 1 ) {

	    va_list ap;
	   	va_start (ap, ftext);
	    vsprintf (TIMER_LogText, ftext, ap);
	    va_end (ap);

		sprintf(TIMER_LogText,"%s\n",TIMER_LogText);

		PrintToTIMEROutput 		(  TIMER_LogText );

	}
//	PrintToEmuOutput (  TIMER_LogText ); 

}



void	TIMERS_Init( void ) {

#ifdef EMU_LOG
	EMU_LOG("[TIMERS]Initialisation");
#endif


	/* Emulator Initialisation */
	Timer[ PS2_TIMER_0 ].Reg = ( TimersReg * ) &SpecialAreaMemory[   ( T0_COUNT & 0xFFFF )  ];
	Timer[ PS2_TIMER_1 ].Reg = ( TimersReg * ) &SpecialAreaMemory[   ( T1_COUNT & 0xFFFF )  ];
	Timer[ PS2_TIMER_2 ].Reg = ( TimersReg * ) &SpecialAreaMemory[   ( T2_COUNT & 0xFFFF )  ];
	Timer[ PS2_TIMER_3 ].Reg = ( TimersReg * ) &SpecialAreaMemory[   ( T3_COUNT & 0xFFFF )  ];

	
	Timer[ NSX2_TIMER_GS ].Reg 		= ( TimersReg * ) &TimerGSReg;
	Timer[ NSX2_TIMER_INPUT ].Reg 	= ( TimersReg * ) &TimerInputReg;
	Timer[ NSX2_MAIN_TIMER ].Reg 	= ( TimersReg * ) &TimerMainReg;
	Timer[ NSX2_THREAD_TIMER ].Reg 	= ( TimersReg * ) &TimerThreadReg;

	MipsCpu.Nb_Cycles 		= 0;
	TIMERS_PreviousCycles 	= 0;


	Timer[ NSX2_TIMER_GS ].Reg->comp 		= 	( 50050 / 240 ) ;
	Timer[ NSX2_TIMER_GS ].Reg->count 		=	0;
	Timer[ NSX2_TIMER_GS ].Reg->hold		= 	0;


	Timer[ NSX2_TIMER_INPUT ].Reg->comp 	= 	50050;
	Timer[ NSX2_TIMER_INPUT ].Reg->count 	=	0;

	Timer[ NSX2_THREAD_TIMER ].Reg->comp 	= 	50000;
	Timer[ NSX2_THREAD_TIMER ].Reg->count 	=	0;

 

}




/* Write Registers*/
void	TIMERS_Write ( _u64 fPhysicalAddress , _u128 fVal , _u8 fDataType ) {

	_u8		boucle;
	_u32	Mask;

	_u64 	RealAdd;

	
	
	RealAdd 		= fPhysicalAddress & 0xFFFF;
	switch ( fDataType ) {

		case DATA_QUAD:	MipsMemoryWriteDW 	( SpecialAreaMemory, fVal.lo, RealAdd );
						MipsMemoryWriteDW 	( SpecialAreaMemory, fVal.hi, RealAdd + 8 );break;

				
		case DATA_DW:	MipsMemoryWriteDW	( SpecialAreaMemory, fVal.lo, RealAdd );break;
				
		case DATA_W:	MipsMemoryWriteW	( SpecialAreaMemory, fVal.lo, RealAdd );break;

		case DATA_HW:	MipsMemoryWriteHW	( SpecialAreaMemory, fVal.lo, RealAdd );break;

		case DATA_B:	MipsMemoryWriteB	( SpecialAreaMemory, fVal.lo, RealAdd );break;
										
		default:		

#ifdef EMU_LOG
	EMU_LOG("MemSpecial:Write DataType Not Handled\n");
#endif


						MipsCpu.halt = 1;break;
	}


	
	switch ( fPhysicalAddress  ) {

		case T0_MODE:	//Timer 0 Mode/Status
#ifdef TIMERS_LOG
			TIMERS_LOG("%8.8lx:[TIMERS][WR]T0_MODE : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
#endif

		if ( ( ( fVal.lo >> 10 ) & 0x1 ) == 0x1 ) CLEAR_MODE_EQUF_TIMER( PS2_TIMER_0 ) ;
		if ( ( ( fVal.lo >> 11 ) & 0x1 ) == 0x1 ) CLEAR_MODE_OVFF_TIMER( PS2_TIMER_0 ) ;

		break;

#ifdef TIMERS_LOG_REG
		case T0_COUNT:	//Timer 0 Counter value
			TIMERS_LOG_REG("%8.8lx:[TIMERS][WR]T0_COUNT : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case T0_COMP:	//Timer 0 Compare value
			TIMERS_LOG_REG("%8.8lx:[TIMERS][WR]T0_COMP : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case T0_HOLD:	//Timer 0 Hold Value
			TIMERS_LOG_REG("%8.8lx:[TIMERS][WR]T0_HOLD : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
#endif


		case T1_MODE:	//Timer 1 Mode/Status
#ifdef TIMERS_LOG
			TIMERS_LOG("%8.8lx:[TIMERS][WR]T1_MODE : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
#endif
		if ( ( ( fVal.lo >> 10 ) & 0x1 ) == 0x1 ) CLEAR_MODE_EQUF_TIMER( PS2_TIMER_1 ) ;
		if ( ( ( fVal.lo >> 11 ) & 0x1 ) == 0x1 ) CLEAR_MODE_OVFF_TIMER( PS2_TIMER_1 ) ;


		break;

#ifdef TIMERS_LOG_REG
		case T1_COUNT:	//Timer 1 Counter value
			TIMERS_LOG_REG("%8.8lx:[TIMERS][WR]T1_COUNT : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case T1_COMP:	//Timer 1 Compare value
			TIMERS_LOG_REG("%8.8lx:[TIMERS][WR]T1_COMP : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case T1_HOLD:	//Timer 1 Hold Value
			TIMERS_LOG_REG("%8.8lx:[TIMERS][WR]T1_HOLD : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;
#endif


		case T2_MODE:	//Timer 2 Mode/Status
#ifdef TIMERS_LOG
			TIMERS_LOG("%8.8lx:[TIMERS][WR]T2_MODE : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
#endif

		if ( ( ( fVal.lo >> 10 ) & 0x1 ) == 0x1 ) CLEAR_MODE_EQUF_TIMER( PS2_TIMER_2 ) ;
		if ( ( ( fVal.lo >> 11 ) & 0x1 ) == 0x1 ) CLEAR_MODE_OVFF_TIMER( PS2_TIMER_2 ) ;


		break;

#ifdef TIMERS_LOG_REG
		case T2_COUNT:	//Timer 2 Counter value
			TIMERS_LOG_REG("%8.8lx:[TIMERS][WR]T2_COUNT : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case T2_COMP:	//Timer 2 Compare value
			TIMERS_LOG("%8.8lx:[TIMERS][WR]T2_COMP : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

#endif

		case T3_MODE:	//Timer 3 Mode/Status
#ifdef TIMERS_LOG
			TIMERS_LOG("%8.8lx:[TIMERS][WR]T3_MODE : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
#endif

		if ( ( ( fVal.lo >> 10 ) & 0x1 ) == 0x1 ) CLEAR_MODE_EQUF_TIMER( PS2_TIMER_3 ) ;
		if ( ( ( fVal.lo >> 11 ) & 0x1 ) == 0x1 ) CLEAR_MODE_OVFF_TIMER( PS2_TIMER_3 ) ;


		break;

#ifdef TIMERS_LOG_REG
		case T3_COUNT:	//Timer 3 Counter value
			TIMERS_LOG_REG("%8.8lx:[TIMERS][WR]T3_COUNT : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

		case T3_COMP:	//Timer 3 Compare value
			TIMERS_LOG_REG("%8.8lx:[TIMERS][WR]T3_COMP : 0x%8.8lx ", (long)MipsCpu.PC,fVal.lo);
		break;

#endif
#ifdef TIMERS_LOG_REG
		default: {
#ifdef EMU_LOG
			EMU_LOG("%8.8lx:[TIMERS][WR]???? : 0x%8.8lx ", (long)MipsCpu.PC,fPhysicalAddress);
#endif
		break;

		}
#endif
	}


}




/* Read from DMA Controller/Channel Registers */
void	TIMERS_Read ( _u64 fPhysicalAddress , _u128 * fVal , _u8 fDataType ) {

	_u64 	RealAdd;

	switch ( fPhysicalAddress  ) {

		case T0_MODE:	//Timer 0 Mode/Status
#ifdef TIMERS_LOG
			TIMERS_LOG("%8.8lx:[TIMERS][RD]T0_MODE : 0x%8.8lx ", (long)MipsCpu.PC,Timer[ PS2_TIMER_0 ].Reg->mode);
#endif
		break;

#ifdef TIMERS_LOG_REG
		case T0_COUNT:	//Timer 0 Counter value
			TIMERS_LOG_REG("%8.8lx:[TIMERS][RD]T0_COUNT : 0x%8.8lx ", (long)MipsCpu.PC,Timer[ PS2_TIMER_0 ].Reg->count);
		break;

		case T0_COMP:	//Timer 0 Compare value
			TIMERS_LOG_REG("%8.8lx:[TIMERS][RD]T0_COMP : 0x%8.8lx ", (long)MipsCpu.PC,Timer[ PS2_TIMER_0 ].Reg->comp);
		break;

		case T0_HOLD:	//Timer 0 Hold Value
			TIMERS_LOG_REG("%8.8lx:[TIMERS][RD]T0_HOLD : 0x%8.8lx ", (long)MipsCpu.PC,Timer[ PS2_TIMER_0 ].Reg->hold);
		break;
#endif


		case T1_MODE:	//Timer 1 Mode/Status
#ifdef TIMERS_LOG
			TIMERS_LOG("%8.8lx:[TIMERS][RD]T1_MODE : 0x%8.8lx ", (long)MipsCpu.PC,Timer[ PS2_TIMER_1 ].Reg->mode);
#endif
		break;

#ifdef TIMERS_LOG_REG
		case T1_COUNT:	//Timer 1 Counter value
			TIMERS_LOG_REG("%8.8lx:[TIMERS][RD]T1_COUNT : 0x%8.8lx ", (long)MipsCpu.PC,Timer[ PS2_TIMER_1 ].Reg->count);
		break;

		case T1_COMP:	//Timer 1 Compare value
			TIMERS_LOG_REG("%8.8lx:[TIMERS][RD]T1_COMP : 0x%8.8lx ", (long)MipsCpu.PC,Timer[ PS2_TIMER_1 ].Reg->comp);
		break;

		case T1_HOLD:	//Timer 1 Hold Value
			TIMERS_LOG_REG("%8.8lx:[TIMERS][RD]T1_HOLD : 0x%8.8lx ", (long)MipsCpu.PC,Timer[ PS2_TIMER_1 ].Reg->hold);
		break;
#endif


		case T2_MODE:	//Timer 2 Mode/Status
#ifdef TIMERS_LOG
			TIMERS_LOG("%8.8lx:[TIMERS][RD]T2_MODE : 0x%8.8lx ", (long)MipsCpu.PC,Timer[ PS2_TIMER_2 ].Reg->mode);
#endif
		break;

#ifdef TIMERS_LOG_REG
		case T2_COUNT:	//Timer 2 Counter value
			TIMERS_LOG_REG("%8.8lx:[TIMERS][RD]T2_COUNT : 0x%8.8lx ", (long)MipsCpu.PC,Timer[ PS2_TIMER_2 ].Reg->count);
		break;

		case T2_COMP:	//Timer 2 Compare value
			TIMERS_LOG("%8.8lx:[TIMERS][RD]T2_COMP : 0x%8.8lx ", (long)MipsCpu.PC,Timer[ PS2_TIMER_2 ].Reg->comp);
		break;

#endif

		case T3_MODE:	//Timer 3 Mode/Status
#ifdef TIMERS_LOG
			TIMERS_LOG("%8.8lx:[TIMERS][RD]T3_MODE : 0x%8.8lx ", (long)MipsCpu.PC,Timer[ PS2_TIMER_3 ].Reg->mode);
#endif
		break;

#ifdef TIMERS_LOG_REG
		case T3_COUNT:	//Timer 3 Counter value
			TIMERS_LOG_REG("%8.8lx:[TIMERS][RD]T3_COUNT : 0x%8.8lx ", (long)MipsCpu.PC,Timer[ PS2_TIMER_3 ].Reg->count);
		break;

		case T3_COMP:	//Timer 3 Compare value
			TIMERS_LOG_REG("%8.8lx:[TIMERS][RD]T3_COMP : 0x%8.8lx ", (long)MipsCpu.PC,Timer[ PS2_TIMER_3 ].Reg->comp);
		break;

#endif

#ifdef TIMERS_LOG_REG
		default: {
#ifdef EMU_LOG
			EMU_LOG("%8.8lx:[TIMERS][RD]???? : 0x%8.8lx ", (long)MipsCpu.PC,fPhysicalAddress);
#endif
		break;
		}
#endif
	}


	RealAdd 		= fPhysicalAddress & 0xFFFF;

	switch ( fDataType ) {

		case DATA_QUAD:	MipsMemoryReadDQ 	( SpecialAreaMemory, fVal, RealAdd );break;
					
		case DATA_DW:	MipsMemoryReadDW	( SpecialAreaMemory, fVal, RealAdd );break;

	    case DATA_HW:	MipsMemoryReadHW	( SpecialAreaMemory, fVal, RealAdd );break;
		
		case DATA_W:	MipsMemoryReadW 	( SpecialAreaMemory, fVal, RealAdd );break;

		case DATA_B:	MipsMemoryReadB		( SpecialAreaMemory, fVal, RealAdd );break;

		default:		EMU_LOG("%8.8lx:[TIMERS][RD][UNDEFINED_TYPE] : 0x%8.8lx ", (long)MipsCpu.PC,fPhysicalAddress);
						MipsCpu.halt = 1;break;
	}



}
 


void TIMERS_Update( void ) {
	_u8 boucle;

	for ( boucle = 0 ; boucle < 4 ; boucle ++ ) TIMERS_PS2_Update ( boucle );

	TIMERS_Update_GS		();
	TIMERS_Update_Input		();
	//TIMERS_Update_Thread	();

	TIMERS_PreviousCycles = MipsCpu.Nb_Cycles;
}




